# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import eventlet

import mock
from oslo_config import cfg
from oslo_utils import timeutils
from oslo_utils import uuidutils
import six

from senlin.common import consts
from senlin.common import exception
from senlin.common import utils as common_utils
from senlin.engine.actions import base as ab
from senlin.engine import cluster as cluster_mod
from senlin.engine import dispatcher
from senlin.engine import environment
from senlin.engine import event as EVENT
from senlin.engine import node as node_mod
from senlin.objects import action as ao
from senlin.objects import cluster_policy as cpo
from senlin.objects import dependency as dobj
from senlin.policies import base as policy_mod
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils
from senlin.tests.unit import fakes

CLUSTER_ID = 'e1cfd82b-dc95-46ad-86e8-37864d7be1cd'
OBJID = '571fffb8-f41c-4cbc-945c-cb2937d76f19'
OWNER_ID = 'c7114713-ee68-409d-ba5d-0560a72a386c'
ACTION_ID = '4c2cead2-fd74-418a-9d12-bd2d9bd7a812'
USER_ID = '3c4d64baadcd437d8dd49054899e73dd'
PROJECT_ID = 'cf7a6ae28dde4f46aa8fe55d318a608f'


class DummyAction(ab.Action):

    def __init__(self, target, action, context, **kwargs):
        super(DummyAction, self).__init__(target, action, context, **kwargs)


class ActionBaseTest(base.SenlinTestCase):

    def setUp(self):
        super(ActionBaseTest, self).setUp()

        self.ctx = utils.dummy_context(project=PROJECT_ID, user_id=USER_ID)
        self.action_values = {
            'name': 'FAKE_NAME',
            'cause': 'FAKE_CAUSE',
            'owner': OWNER_ID,
            'interval': 60,
            'start_time': 0,
            'end_time': 0,
            'timeout': 120,
            'status': 'FAKE_STATUS',
            'status_reason': 'FAKE_STATUS_REASON',
            'inputs': {'param': 'value'},
            'outputs': {'key': 'output_value'},
            'created_at': timeutils.utcnow(True),
            'updated_at': None,
            'data': {'data_key': 'data_value'},
        }

    def _verify_new_action(self, obj, target, action):
        self.assertIsNone(obj.id)
        self.assertEqual('', obj.name)
        self.assertEqual(target, obj.target)
        self.assertEqual(action, obj.action)
        self.assertEqual('', obj.cause)
        self.assertIsNone(obj.owner)
        self.assertEqual(-1, obj.interval)
        self.assertIsNone(obj.start_time)
        self.assertIsNone(obj.end_time)
        self.assertEqual(cfg.CONF.default_action_timeout, obj.timeout)
        self.assertEqual('INIT', obj.status)
        self.assertEqual('', obj.status_reason)
        self.assertEqual({}, obj.inputs)
        self.assertEqual({}, obj.outputs)
        self.assertIsNone(obj.created_at)
        self.assertIsNone(obj.updated_at)
        self.assertEqual({}, obj.data)

    @mock.patch.object(cluster_mod.Cluster, 'load')
    def test_action_new_cluster(self, mock_load):
        fake_cluster = mock.Mock(timeout=cfg.CONF.default_action_timeout)
        mock_load.return_value = fake_cluster
        obj = ab.Action(OBJID, 'CLUSTER_CREATE', self.ctx)
        self._verify_new_action(obj, OBJID, 'CLUSTER_CREATE')

    @mock.patch.object(node_mod.Node, 'load')
    def test_action_new_node(self, mock_load):
        obj = ab.Action(OBJID, 'NODE_CREATE', self.ctx)
        self._verify_new_action(obj, OBJID, 'NODE_CREATE')

    def test_action_init_with_values(self):
        values = copy.deepcopy(self.action_values)
        values['id'] = 'FAKE_ID'
        values['created_at'] = 'FAKE_CREATED_TIME'
        values['updated_at'] = 'FAKE_UPDATED_TIME'

        obj = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)

        self.assertEqual('FAKE_ID', obj.id)
        self.assertEqual('FAKE_NAME', obj.name)
        self.assertEqual(OBJID, obj.target)
        self.assertEqual('FAKE_CAUSE', obj.cause)
        self.assertEqual(OWNER_ID, obj.owner)
        self.assertEqual(60, obj.interval)
        self.assertEqual(0, obj.start_time)
        self.assertEqual(0, obj.end_time)
        self.assertEqual(120, obj.timeout)
        self.assertEqual('FAKE_STATUS', obj.status)
        self.assertEqual('FAKE_STATUS_REASON', obj.status_reason)
        self.assertEqual({'param': 'value'}, obj.inputs)
        self.assertEqual({'key': 'output_value'}, obj.outputs)
        self.assertEqual('FAKE_CREATED_TIME', obj.created_at)
        self.assertEqual('FAKE_UPDATED_TIME', obj.updated_at)
        self.assertEqual({'data_key': 'data_value'}, obj.data)

    def test_action_store_for_create(self):
        values = copy.deepcopy(self.action_values)
        obj = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)

        self.assertEqual(common_utils.isotime(values['created_at']),
                         common_utils.isotime(obj.created_at))
        self.assertIsNone(obj.updated_at)

        # store for creation
        res = obj.store(self.ctx)
        self.assertIsNotNone(res)
        self.assertEqual(obj.id, res)
        self.assertIsNotNone(obj.created_at)
        self.assertIsNone(obj.updated_at)

    def test_action_store_for_update(self):
        values = copy.deepcopy(self.action_values)

        obj = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)
        obj_id = obj.store(self.ctx)
        self.assertIsNotNone(obj_id)
        self.assertIsNotNone(obj.created_at)
        self.assertIsNone(obj.updated_at)

        # store for creation
        res = obj.store(self.ctx)
        self.assertIsNotNone(res)
        self.assertEqual(obj_id, res)
        self.assertEqual(obj.id, res)
        self.assertIsNotNone(obj.created_at)
        self.assertIsNotNone(obj.updated_at)

    def test_from_db_record(self):
        values = copy.deepcopy(self.action_values)
        obj = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)
        obj.store(self.ctx)

        record = ao.Action.get(self.ctx, obj.id)

        action_obj = ab.Action._from_object(record)
        self.assertIsInstance(action_obj, ab.Action)
        self.assertEqual(obj.id, action_obj.id)
        self.assertEqual(obj.action, action_obj.action)
        self.assertEqual(obj.name, action_obj.name)
        self.assertEqual(obj.target, action_obj.target)
        self.assertEqual(obj.cause, action_obj.cause)
        self.assertEqual(obj.owner, action_obj.owner)
        self.assertEqual(obj.interval, action_obj.interval)
        self.assertEqual(obj.start_time, action_obj.start_time)
        self.assertEqual(obj.end_time, action_obj.end_time)
        self.assertEqual(obj.timeout, action_obj.timeout)
        self.assertEqual(obj.status, action_obj.status)
        self.assertEqual(obj.status_reason, action_obj.status_reason)
        self.assertEqual(obj.inputs, action_obj.inputs)
        self.assertEqual(obj.outputs, action_obj.outputs)
        self.assertEqual(common_utils.isotime(obj.created_at),
                         common_utils.isotime(action_obj.created_at))
        self.assertEqual(obj.updated_at, action_obj.updated_at)
        self.assertEqual(obj.data, action_obj.data)
        self.assertEqual(obj.user, action_obj.user)
        self.assertEqual(obj.project, action_obj.project)
        self.assertEqual(obj.domain, action_obj.domain)

    def test_from_db_record_with_empty_fields(self):
        values = copy.deepcopy(self.action_values)
        del values['inputs']
        del values['outputs']
        obj = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)
        obj.store(self.ctx)
        record = ao.Action.get(self.ctx, obj.id)
        action_obj = ab.Action._from_object(record)
        self.assertEqual({}, action_obj.inputs)
        self.assertEqual({}, action_obj.outputs)

    def test_load(self):
        values = copy.deepcopy(self.action_values)
        obj = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)
        obj.store(self.ctx)

        result = ab.Action.load(self.ctx, obj.id, None)
        # no need to do a thorough test here
        self.assertEqual(obj.id, result.id)
        self.assertEqual(obj.action, result.action)

        db_action = ao.Action.get(self.ctx, obj.id)
        result = ab.Action.load(self.ctx, None, db_action)
        # no need to do a thorough test here
        self.assertEqual(obj.id, result.id)
        self.assertEqual(obj.action, result.action)

    def test_load_not_found(self):
        # not found due to bad identity
        ex = self.assertRaises(exception.ResourceNotFound,
                               ab.Action.load,
                               self.ctx, 'non-existent', None)
        self.assertEqual("The action 'non-existent' could not be "
                         "found.", six.text_type(ex))

        # not found due to no object
        self.patchobject(ao.Action, 'get', return_value=None)
        ex = self.assertRaises(exception.ResourceNotFound,
                               ab.Action.load,
                               self.ctx, 'whatever', None)
        self.assertEqual("The action 'whatever' could not be found.",
                         six.text_type(ex))

    @mock.patch.object(ab.Action, 'store')
    def test_action_create(self, mock_store):
        mock_store.return_value = 'FAKE_ID'

        result = ab.Action.create(self.ctx, OBJID, 'CLUSTER_DANCE',
                                  name='test')

        self.assertEqual('FAKE_ID', result)
        mock_store.assert_called_once_with(self.ctx)

    def test_action_delete(self):
        result = ab.Action.delete(self.ctx, 'non-existent')
        self.assertIsNone(result)

        values = copy.deepcopy(self.action_values)
        action1 = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)
        action1.store(self.ctx)

        result = ab.Action.delete(self.ctx, action1.id)
        self.assertIsNone(result)

    @mock.patch.object(ao.Action, 'delete')
    def test_action_delete_db_call(self, mock_call):
        # test db api call
        ab.Action.delete(self.ctx, 'FAKE_ID')
        mock_call.assert_called_once_with(self.ctx, 'FAKE_ID')

    @mock.patch.object(ao.Action, 'signal')
    def test_action_signal_bad_command(self, mock_call):
        values = copy.deepcopy(self.action_values)
        action1 = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)
        action1.store(self.ctx)

        result = action1.signal('BOGUS')
        self.assertIsNone(result)
        self.assertEqual(0, mock_call.call_count)

    @mock.patch.object(ao.Action, 'signal')
    def test_action_signal_cancel(self, mock_call):
        values = copy.deepcopy(self.action_values)
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, **values)
        action.store(self.ctx)

        expected = [action.INIT, action.WAITING, action.READY, action.RUNNING]
        for status in expected:
            action.status = status
            result = action.signal(action.SIG_CANCEL)
            self.assertIsNone(result)
            self.assertEqual(1, mock_call.call_count)
            mock_call.reset_mock()

        invalid = [action.SUSPENDED, action.SUCCEEDED, action.CANCELLED,
                   action.FAILED]
        for status in invalid:
            action.status = status
            result = action.signal(action.SIG_CANCEL)
            self.assertIsNone(result)
            self.assertEqual(0, mock_call.call_count)
            mock_call.reset_mock()

    @mock.patch.object(ao.Action, 'signal')
    def test_action_signal_suspend(self, mock_call):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, id=ACTION_ID)

        expected = [action.RUNNING]
        for status in expected:
            action.status = status
            result = action.signal(action.SIG_SUSPEND)
            self.assertIsNone(result)
            self.assertEqual(1, mock_call.call_count)
            mock_call.reset_mock()

        invalid = [action.INIT, action.WAITING, action.READY, action.SUSPENDED,
                   action.SUCCEEDED, action.CANCELLED, action.FAILED]
        for status in invalid:
            action.status = status
            result = action.signal(action.SIG_SUSPEND)
            self.assertIsNone(result)
            self.assertEqual(0, mock_call.call_count)
            mock_call.reset_mock()

    @mock.patch.object(ao.Action, 'signal')
    def test_action_signal_resume(self, mock_call):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, id=ACTION_ID)

        expected = [action.SUSPENDED]
        for status in expected:
            action.status = status
            result = action.signal(action.SIG_RESUME)
            self.assertIsNone(result)
            self.assertEqual(1, mock_call.call_count)
            mock_call.reset_mock()

        invalid = [action.INIT, action.WAITING, action.READY, action.RUNNING,
                   action.SUCCEEDED, action.CANCELLED, action.FAILED]
        for status in invalid:
            action.status = status
            result = action.signal(action.SIG_RESUME)
            self.assertIsNone(result)
            self.assertEqual(0, mock_call.call_count)
            mock_call.reset_mock()

    def test_execute_default(self):
        action = ab.Action.__new__(DummyAction, OBJID, 'BOOM', self.ctx)
        self.assertRaises(NotImplementedError,
                          action.execute)

    @mock.patch.object(EVENT, 'info')
    @mock.patch.object(EVENT, 'error')
    @mock.patch.object(EVENT, 'warning')
    @mock.patch.object(ao.Action, 'mark_succeeded')
    @mock.patch.object(ao.Action, 'mark_failed')
    @mock.patch.object(ao.Action, 'mark_cancelled')
    @mock.patch.object(ao.Action, 'mark_ready')
    @mock.patch.object(ao.Action, 'abandon')
    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(eventlet, 'sleep')
    def test_set_status(self, mock_sleep, mock_start, mock_abandon,
                        mark_ready, mark_cancel, mark_fail,
                        mark_succeed, mock_event, mock_error,
                        mock_info):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, id='FAKE_ID')
        action.entity = mock.Mock()

        action.set_status(action.RES_OK, 'FAKE_REASON')

        self.assertEqual(action.SUCCEEDED, action.status)
        self.assertEqual('FAKE_REASON', action.status_reason)
        mark_succeed.assert_called_once_with(action.context, 'FAKE_ID',
                                             mock.ANY)

        action.set_status(action.RES_ERROR, 'FAKE_ERROR')
        self.assertEqual(action.FAILED, action.status)
        self.assertEqual('FAKE_ERROR', action.status_reason)
        mark_fail.assert_called_once_with(action.context, 'FAKE_ID', mock.ANY,
                                          'FAKE_ERROR')

        mark_fail.reset_mock()
        action.set_status(action.RES_TIMEOUT, 'TIMEOUT_ERROR')
        self.assertEqual(action.FAILED, action.status)
        self.assertEqual('TIMEOUT_ERROR', action.status_reason)
        mark_fail.assert_called_once_with(action.context, 'FAKE_ID', mock.ANY,
                                          'TIMEOUT_ERROR')

        mark_fail.reset_mock()
        action.set_status(action.RES_CANCEL, 'CANCELLED')
        self.assertEqual(action.CANCELLED, action.status)
        self.assertEqual('CANCELLED', action.status_reason)
        mark_cancel.assert_called_once_with(action.context, 'FAKE_ID',
                                            mock.ANY)

        mark_fail.reset_mock()
        action.set_status(action.RES_LIFECYCLE_COMPLETE, 'LIFECYCLE COMPLETE')
        self.assertEqual(action.SUCCEEDED, action.status)
        self.assertEqual('LIFECYCLE COMPLETE', action.status_reason)
        mark_ready.assert_called_once_with(action.context, 'FAKE_ID', mock.ANY)

        mark_fail.reset_mock()
        action.set_status(action.RES_RETRY, 'BUSY')
        self.assertEqual(action.READY, action.status)
        self.assertEqual('BUSY', action.status_reason)
        mock_start.assert_called_once_with(action.id)
        mock_sleep.assert_called_once_with(10)
        mock_abandon.assert_called_once_with(
            action.context, 'FAKE_ID', {'data': {'retries': 1}})

        mark_fail.reset_mock()
        action.data = {'retries': 3}
        action.set_status(action.RES_RETRY, 'BUSY')
        self.assertEqual(action.RES_ERROR, action.status)
        mark_fail.assert_called_once_with(action.context, 'FAKE_ID', mock.ANY,
                                          'BUSY')

    @mock.patch.object(EVENT, 'info')
    @mock.patch.object(EVENT, 'error')
    @mock.patch.object(EVENT, 'warning')
    @mock.patch.object(ao.Action, 'mark_succeeded')
    @mock.patch.object(ao.Action, 'mark_failed')
    @mock.patch.object(ao.Action, 'abandon')
    def test_set_status_dump_event(self, mock_abandon, mark_fail,
                                   mark_succeed, mock_warning, mock_error,
                                   mock_info):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, id='FAKE_ID')
        action.entity = mock.Mock()

        action.set_status(action.RES_OK, 'FAKE_SUCCEEDED')
        mock_info.assert_called_once_with(action, consts.PHASE_END,
                                          'FAKE_SUCCEEDED')

        action.set_status(action.RES_ERROR, 'FAKE_ERROR')
        mock_error.assert_called_once_with(action, consts.PHASE_ERROR,
                                           'FAKE_ERROR')

        action.set_status(action.RES_RETRY, 'FAKE_RETRY')
        mock_warning.assert_called_once_with(action, consts.PHASE_ERROR,
                                             'FAKE_RETRY')

    @mock.patch.object(EVENT, 'info')
    @mock.patch.object(EVENT, 'error')
    @mock.patch.object(EVENT, 'warning')
    @mock.patch.object(ao.Action, 'mark_succeeded')
    @mock.patch.object(ao.Action, 'mark_failed')
    @mock.patch.object(ao.Action, 'abandon')
    def test_set_status_reason_is_none(self, mock_abandon, mark_fail,
                                       mark_succeed, mock_warning, mock_error,
                                       mock_info):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, id='FAKE_ID')
        action.entity = mock.Mock()

        action.set_status(action.RES_OK)
        mock_info.assert_called_once_with(action, consts.PHASE_END,
                                          'SUCCEEDED')

        action.set_status(action.RES_ERROR)
        mock_error.assert_called_once_with(action, consts.PHASE_ERROR,
                                           'ERROR')

        action.set_status(action.RES_RETRY)
        mock_warning.assert_called_once_with(action, consts.PHASE_ERROR,
                                             'RETRY')

    @mock.patch.object(ao.Action, 'check_status')
    def test_get_status(self, mock_get):
        mock_get.return_value = 'FAKE_STATUS'

        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        action.id = 'FAKE_ID'

        res = action.get_status()

        self.assertEqual('FAKE_STATUS', res)
        self.assertEqual('FAKE_STATUS', action.status)
        mock_get.assert_called_once_with(action.context, 'FAKE_ID', mock.ANY)

    @mock.patch.object(ab, 'wallclock')
    def test_is_timeout(self, mock_time):
        action = ab.Action.__new__(DummyAction, 'OBJ', 'BOOM', self.ctx)
        action.start_time = 1
        action.timeout = 10

        mock_time.return_value = 9
        self.assertFalse(action.is_timeout())

        mock_time.return_value = 10
        self.assertFalse(action.is_timeout())

        mock_time.return_value = 11
        self.assertFalse(action.is_timeout())

        mock_time.return_value = 12
        self.assertTrue(action.is_timeout())

    @mock.patch.object(EVENT, 'debug')
    def test_check_signal_timeout(self, mock_debug):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx, id='FAKE_ID',
                           timeout=10)
        action.entity = mock.Mock()
        self.patchobject(action, 'is_timeout', return_value=True)

        res = action._check_signal()
        self.assertEqual(action.RES_TIMEOUT, res)

    @mock.patch.object(ao.Action, 'signal_query')
    def test_check_signal_signals_caught(self, mock_query):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        action.id = 'FAKE_ID'
        action.timeout = 100
        self.patchobject(action, 'is_timeout', return_value=False)
        sig_cmd = mock.Mock()
        mock_query.return_value = sig_cmd

        res = action._check_signal()
        self.assertEqual(sig_cmd, res)
        mock_query.assert_called_once_with(action.context, 'FAKE_ID')

    @mock.patch.object(ao.Action, 'signal_query')
    def test_is_cancelled(self, mock_query):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        action.id = 'FAKE_ID'
        action.timeout = 100
        self.patchobject(action, 'is_timeout', return_value=False)

        mock_query.return_value = action.SIG_CANCEL
        res = action.is_cancelled()
        self.assertTrue(res)
        mock_query.assert_called_once_with(action.context, 'FAKE_ID')
        mock_query.reset_mock()

        mock_query.return_value = None
        res = action.is_cancelled()
        self.assertFalse(res)
        mock_query.assert_called_once_with(action.context, 'FAKE_ID')

    @mock.patch.object(ao.Action, 'signal_query')
    def test_is_suspended(self, mock_query):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        action.id = 'FAKE_ID'
        action.timeout = 100
        self.patchobject(action, 'is_timeout', return_value=False)

        mock_query.return_value = action.SIG_SUSPEND
        res = action.is_suspended()
        self.assertTrue(res)
        mock_query.assert_called_once_with(action.context, 'FAKE_ID')
        mock_query.reset_mock()

        mock_query.return_value = 'OTHERS'
        res = action.is_suspended()
        self.assertFalse(res)
        mock_query.assert_called_once_with(action.context, 'FAKE_ID')

    @mock.patch.object(ao.Action, 'signal_query')
    def test_is_resumed(self, mock_query):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        action.id = 'FAKE_ID'
        action.timeout = 100
        self.patchobject(action, 'is_timeout', return_value=False)

        mock_query.return_value = action.SIG_RESUME
        res = action.is_resumed()
        self.assertTrue(res)
        mock_query.assert_called_once_with(action.context, 'FAKE_ID')
        mock_query.reset_mock()

        mock_query.return_value = 'OTHERS'
        res = action.is_resumed()
        self.assertFalse(res)
        mock_query.assert_called_once_with(action.context, 'FAKE_ID')

    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    def test_policy_check_target_invalid(self, mock_load):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)

        res = action.policy_check('FAKE_CLUSTER', 'WHEN')

        self.assertIsNone(res)
        self.assertEqual(0, mock_load.call_count)

    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    def test_policy_check_no_bindings(self, mock_load):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        mock_load.return_value = []

        res = action.policy_check('FAKE_CLUSTER', 'BEFORE')

        self.assertIsNone(res)
        self.assertEqual(policy_mod.CHECK_OK, action.data['status'])
        mock_load.assert_called_once_with(action.context, 'FAKE_CLUSTER',
                                          sort='priority',
                                          filters={'enabled': True})

    @mock.patch.object(dobj.Dependency, 'get_depended')
    @mock.patch.object(dobj.Dependency, 'get_dependents')
    def test_action_to_dict(self, mock_dep_by, mock_dep_on):
        mock_dep_on.return_value = ['ACTION_1']
        mock_dep_by.return_value = ['ACTION_2']
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx,
                           **self.action_values)
        action.id = 'FAKE_ID'
        ts = common_utils.isotime(self.action_values['created_at'])
        expected = {
            'id': 'FAKE_ID',
            'name': 'FAKE_NAME',
            'action': 'OBJECT_ACTION',
            'target': OBJID,
            'cause': 'FAKE_CAUSE',
            'owner': OWNER_ID,
            'interval': 60,
            'start_time': 0,
            'end_time': 0,
            'timeout': 120,
            'status': 'FAKE_STATUS',
            'status_reason': 'FAKE_STATUS_REASON',
            'inputs': {'param': 'value'},
            'outputs': {'key': 'output_value'},
            'depends_on': ['ACTION_1'],
            'depended_by': ['ACTION_2'],
            'created_at': ts,
            'updated_at': None,
            'data': {'data_key': 'data_value'},
            'user': USER_ID,
            'project': PROJECT_ID,
        }

        res = action.to_dict()
        self.assertEqual(expected, res)
        mock_dep_on.assert_called_once_with(action.context, 'FAKE_ID')
        mock_dep_by.assert_called_once_with(action.context, 'FAKE_ID')


class ActionPolicyCheckTest(base.SenlinTestCase):

    def setUp(self):
        super(ActionPolicyCheckTest, self).setUp()

        self.ctx = utils.dummy_context()
        environment.global_env().register_policy('DummyPolicy',
                                                 fakes.TestPolicy)

    def _create_policy(self):
        values = {
            'user': self.ctx.user_id,
            'project': self.ctx.project_id,
        }
        policy = fakes.TestPolicy('DummyPolicy', 'test-policy', **values)
        policy.store(self.ctx)
        return policy

    def _create_cp_binding(self, cluster_id, policy_id):
        return cpo.ClusterPolicy(cluster_id=cluster_id, policy_id=policy_id,
                                 enabled=True, id=uuidutils.generate_uuid(),
                                 last_op=None)

    @mock.patch.object(policy_mod.Policy, 'post_op')
    @mock.patch.object(policy_mod.Policy, 'pre_op')
    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(policy_mod.Policy, 'load')
    def test_policy_check_missing_target(self, mock_load, mock_load_all,
                                         mock_pre_op, mock_post_op):
        cluster_id = CLUSTER_ID
        # Note: policy is mocked
        spec = {
            'type': 'TestPolicy',
            'version': '1.0',
            'properties': {'KEY2': 5},
        }
        policy = fakes.TestPolicy('test-policy', spec)
        policy.id = uuidutils.generate_uuid()
        policy.TARGET = [('BEFORE', 'OBJECT_ACTION')]
        # Note: policy binding is created but not stored
        pb = self._create_cp_binding(cluster_id, policy.id)
        self.assertIsNone(pb.last_op)
        mock_load_all.return_value = [pb]
        mock_load.return_value = policy
        mock_pre_op.return_value = None
        mock_post_op.return_value = None
        action = ab.Action(cluster_id, 'OBJECT_ACTION_1', self.ctx)

        res = action.policy_check(cluster_id, 'AFTER')

        self.assertIsNone(res)
        self.assertEqual(policy_mod.CHECK_OK, action.data['status'])
        mock_load_all.assert_called_once_with(
            action.context, cluster_id, sort='priority',
            filters={'enabled': True})
        mock_load.assert_called_once_with(action.context, policy.id)
        # last_op was updated anyway
        self.assertIsNotNone(pb.last_op)
        # neither pre_op nor post_op was called, because target not match
        self.assertEqual(0, mock_pre_op.call_count)
        self.assertEqual(0, mock_post_op.call_count)

    def test__check_result_true(self):
        cluster_id = CLUSTER_ID
        action = ab.Action(cluster_id, 'OBJECT_ACTION', self.ctx)
        action.data['status'] = policy_mod.CHECK_OK
        action.data['reason'] = "Completed policy checking."

        res = action._check_result('FAKE_POLICY_NAME')

        self.assertTrue(res)

    def test__check_result_false(self):
        cluster_id = CLUSTER_ID
        action = ab.Action(cluster_id, 'OBJECT_ACTION', self.ctx)
        action.data['status'] = policy_mod.CHECK_ERROR
        reason = ("Policy '%s' cooldown is still in progress." %
                  'FAKE_POLICY_2')
        action.data['reason'] = reason

        res = action._check_result('FAKE_POLICY_NAME')
        reason = ("Failed policy '%(name)s': %(reason)s"
                  ) % {'name': 'FAKE_POLICY_NAME', 'reason': reason}
        self.assertFalse(res)

    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(policy_mod.Policy, 'load')
    def test_policy_check_pre_op(self, mock_load, mock_load_all):
        cluster_id = CLUSTER_ID
        # Note: policy is mocked
        spec = {
            'type': 'TestPolicy',
            'version': '1.0',
            'properties': {'KEY2': 5},
        }
        policy = fakes.TestPolicy('test-policy', spec)
        policy.id = uuidutils.generate_uuid()
        policy.TARGET = [('BEFORE', 'OBJECT_ACTION')]
        # Note: policy binding is created but not stored
        pb = self._create_cp_binding(cluster_id, policy.id)
        self.assertIsNone(pb.last_op)
        mock_load_all.return_value = [pb]
        mock_load.return_value = policy
        entity = mock.Mock()
        action = ab.Action(cluster_id, 'OBJECT_ACTION', self.ctx)
        action.entity = entity

        res = action.policy_check(cluster_id, 'BEFORE')

        self.assertIsNone(res)
        self.assertEqual(policy_mod.CHECK_OK, action.data['status'])
        mock_load_all.assert_called_once_with(
            action.context, cluster_id, sort='priority',
            filters={'enabled': True})
        mock_load.assert_called_once_with(action.context, policy.id)
        # last_op was not updated
        self.assertIsNone(pb.last_op)

    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(policy_mod.Policy, 'load')
    def test_policy_check_post_op(self, mock_load, mock_load_all):
        cluster_id = CLUSTER_ID
        # Note: policy is mocked
        policy = mock.Mock(id=uuidutils.generate_uuid(), cooldown=0,
                           TARGET=[('AFTER', 'OBJECT_ACTION')])
        # Note: policy binding is created but not stored
        pb = self._create_cp_binding(cluster_id, policy.id)
        self.assertIsNone(pb.last_op)
        mock_load_all.return_value = [pb]
        mock_load.return_value = policy
        entity = mock.Mock()
        action = ab.Action(cluster_id, 'OBJECT_ACTION', self.ctx)
        action.entity = entity

        res = action.policy_check(CLUSTER_ID, 'AFTER')

        self.assertIsNone(res)
        self.assertEqual(policy_mod.CHECK_OK, action.data['status'])
        mock_load_all.assert_called_once_with(
            action.context, cluster_id, sort='priority',
            filters={'enabled': True})
        mock_load.assert_called_once_with(action.context, policy.id)
        # last_op was updated for POST check
        self.assertIsNotNone(pb.last_op)
        # pre_op is called, but post_op was not called
        self.assertEqual(0, policy.pre_op.call_count)
        policy.post_op.assert_called_once_with(cluster_id, action)

    @mock.patch.object(cpo.ClusterPolicy, 'cooldown_inprogress')
    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(policy_mod.Policy, 'load')
    def test_policy_check_cooldown_inprogress(self, mock_load, mock_load_all,
                                              mock_inprogress):
        cluster_id = CLUSTER_ID
        # Note: policy is mocked
        policy_id = uuidutils.generate_uuid()
        policy = mock.Mock(id=policy_id, TARGET=[('AFTER', 'OBJECT_ACTION')])
        # Note: policy binding is created but not stored
        pb = self._create_cp_binding(cluster_id, policy.id)
        mock_inprogress.return_value = True
        mock_load_all.return_value = [pb]
        mock_load.return_value = policy
        action = ab.Action(cluster_id, 'OBJECT_ACTION', self.ctx)

        # Do it
        res = action.policy_check(CLUSTER_ID, 'AFTER')

        self.assertIsNone(res)
        self.assertEqual(policy_mod.CHECK_ERROR, action.data['status'])
        self.assertEqual(
            'Policy %s cooldown is still in progress.' % policy_id,
            six.text_type(action.data['reason']))
        mock_load_all.assert_called_once_with(
            action.context, cluster_id, sort='priority',
            filters={'enabled': True})
        mock_load.assert_called_once_with(action.context, policy.id)
        # last_op was updated for POST check
        self.assertIsNotNone(pb.last_op)
        # neither pre_op nor post_op was not called, due to cooldown
        self.assertEqual(0, policy.pre_op.call_count)
        self.assertEqual(0, policy.post_op.call_count)

    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(policy_mod.Policy, 'load')
    @mock.patch.object(ab.Action, '_check_result')
    def test_policy_check_abort_in_middle(self, mock_check, mock_load,
                                          mock_load_all):
        cluster_id = CLUSTER_ID
        # Note: both policies are mocked
        policy1 = mock.Mock(id=uuidutils.generate_uuid(), cooldown=0,
                            TARGET=[('AFTER', 'OBJECT_ACTION')])
        policy1.name = 'P1'
        policy2 = mock.Mock(id=uuidutils.generate_uuid(), cooldown=0,
                            TARGET=[('AFTER', 'OBJECT_ACTION')])
        policy2.name = 'P2'
        action = ab.Action(cluster_id, 'OBJECT_ACTION', self.ctx)

        # Note: policy binding is created but not stored
        pb1 = self._create_cp_binding(cluster_id, policy1.id)
        pb2 = self._create_cp_binding(cluster_id, policy2.id)
        mock_load_all.return_value = [pb1, pb2]
        # mock return value for two calls
        mock_load.side_effect = [policy1, policy2]
        mock_check.side_effect = [False, True]

        res = action.policy_check(cluster_id, 'AFTER')

        self.assertIsNone(res)

        # post_op from policy1 was called, but post_op from policy2 was not
        policy1.post_op.assert_called_once_with(cluster_id, action)
        self.assertEqual(0, policy2.post_op.call_count)

        mock_load_all.assert_called_once_with(
            action.context, cluster_id, sort='priority',
            filters={'enabled': True})
        calls = [mock.call(action.context, policy1.id)]
        mock_load.assert_has_calls(calls)


class ActionProcTest(base.SenlinTestCase):

    def setUp(self):
        super(ActionProcTest, self).setUp()

        self.ctx = utils.dummy_context()

    @mock.patch.object(EVENT, 'info')
    @mock.patch.object(ab.Action, 'load')
    @mock.patch.object(ao.Action, 'mark_succeeded')
    def test_action_proc_successful(self, mock_mark, mock_load,
                                    mock_event_info):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        mock_obj = mock.Mock()
        action.entity = mock_obj
        self.patchobject(action, 'execute',
                         return_value=(action.RES_OK, 'BIG SUCCESS'))
        mock_status = self.patchobject(action, 'set_status')
        mock_load.return_value = action

        res = ab.ActionProc(self.ctx, 'ACTION_ID')

        self.assertTrue(res)
        mock_load.assert_called_once_with(self.ctx, action_id='ACTION_ID',
                                          project_safe=False)
        mock_event_info.assert_called_once_with(action, 'start', 'ACTION_I')
        mock_status.assert_called_once_with(action.RES_OK, 'BIG SUCCESS')

    @mock.patch.object(EVENT, 'info')
    @mock.patch.object(ab.Action, 'load')
    @mock.patch.object(ao.Action, 'mark_failed')
    def test_action_proc_failed_error(self, mock_mark, mock_load, mock_info):
        action = ab.Action(OBJID, 'CLUSTER_ACTION', self.ctx, id=ACTION_ID)
        action.entity = mock.Mock(id=CLUSTER_ID, name='fake-cluster')

        self.patchobject(action, 'execute', side_effect=Exception('Boom!'))
        mock_status = self.patchobject(action, 'set_status')
        mock_load.return_value = action

        res = ab.ActionProc(self.ctx, 'ACTION')

        self.assertFalse(res)
        mock_load.assert_called_once_with(self.ctx, action_id='ACTION',
                                          project_safe=False)
        mock_info.assert_called_once_with(action, 'start', 'ACTION')
        mock_status.assert_called_once_with(action.RES_ERROR, 'Boom!')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.engine.actions import base as ab
from senlin.engine.actions import cluster_action as ca
from senlin.engine import cluster as cm
from senlin.engine import scheduler
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class ClusterActionWaitTest(base.SenlinTestCase):
    scenarios = [
        ('wait_ready', dict(
            statuses=[
                ab.Action.WAITING,
                ab.Action.READY
            ],
            cancelled=[False, False],
            timeout=[False, False],
            failed=[False, False],
            code=ab.Action.RES_OK,
            rescheduled_times=1,
            message='All dependents ended with success')
         ),
        ('wait_fail', dict(
            statuses=[
                ab.Action.WAITING,
                ab.Action.FAILED
            ],
            cancelled=[False, False],
            timeout=[False, False],
            code=ab.Action.RES_ERROR,
            rescheduled_times=1,
            message='ACTION [FAKE_ID] failed')
         ),
        ('wait_wait_cancel', dict(
            statuses=[
                ab.Action.WAITING,
                ab.Action.WAITING,
                ab.Action.WAITING,
            ],
            cancelled=[False, False, True],
            timeout=[False, False, False],
            code=ab.Action.RES_CANCEL,
            rescheduled_times=2,
            message='ACTION [FAKE_ID] cancelled')
         ),
        ('wait_wait_timeout', dict(
            statuses=[
                ab.Action.WAITING,
                ab.Action.WAITING,
                ab.Action.WAITING,
            ],
            cancelled=[False, False, False],
            timeout=[False, False, True],
            code=ab.Action.RES_TIMEOUT,
            rescheduled_times=2,
            message='ACTION [FAKE_ID] timeout')
         ),

    ]

    def setUp(self):
        super(ClusterActionWaitTest, self).setUp()
        self.ctx = utils.dummy_context()

    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(scheduler, 'reschedule')
    def test_wait_dependents(self, mock_reschedule, mock_load):
        action = ca.ClusterAction('ID', 'ACTION', self.ctx)
        action.id = 'FAKE_ID'
        self.patchobject(action, 'get_status', side_effect=self.statuses)
        self.patchobject(action, 'is_cancelled', side_effect=self.cancelled)
        self.patchobject(action, 'is_timeout', side_effect=self.timeout)

        res_code, res_msg = action._wait_for_dependents()
        self.assertEqual(self.code, res_code)
        self.assertEqual(self.message, res_msg)
        self.assertEqual(self.rescheduled_times, mock_reschedule.call_count)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.common import consts
from senlin.common import scaleutils
from senlin.engine.actions import cluster_action as ca
from senlin.engine import cluster as cm
from senlin.objects import node as no
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


@mock.patch.object(cm.Cluster, 'load')
class ClusterScaleInTest(base.SenlinTestCase):

    def setUp(self):
        super(ClusterScaleInTest, self).setUp()
        self.ctx = utils.dummy_context()

    @mock.patch.object(scaleutils, 'nodes_by_random')
    @mock.patch.object(ca.ClusterAction, '_delete_nodes')
    @mock.patch.object(no.Node, 'count_by_cluster')
    def test_do_scale_in_no_pd_no_count(self, mock_count, mock_delete,
                                        mock_select, mock_load):
        cluster = mock.Mock(id='CID', min_size=1, max_size=-1)
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_ACTION', self.ctx,
                                  data={}, inputs={})
        mock_count.return_value = 10
        mock_delete.return_value = (action.RES_OK, 'Life is beautiful.')

        # do it
        res_code, res_msg = action.do_scale_in()

        # assertions
        self.assertEqual(action.RES_OK, res_code)
        self.assertEqual('Cluster scaling succeeded.', res_msg)

        # deleting 1 nodes
        mock_count.assert_called_once_with(action.context, 'CID')
        mock_delete.assert_called_once_with(mock.ANY)
        mock_select.assert_called_once_with(cluster.nodes, 1)
        cluster.set_status.assert_called_once_with(
            action.context, consts.CS_RESIZING, 'Cluster scale in started.',
            desired_capacity=9)
        cluster.eval_status.assert_called_once_with(
            action.context, consts.CLUSTER_SCALE_IN)

    @mock.patch.object(ca.ClusterAction, '_sleep')
    @mock.patch.object(ca.ClusterAction, '_delete_nodes')
    @mock.patch.object(no.Node, 'count_by_cluster')
    def test_do_scale_in_with_pd_no_input(self, mock_count, mock_delete,
                                          mock_sleep, mock_load):
        cluster = mock.Mock(id='CID', min_size=1, max_size=-1)
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_ACTION', self.ctx)
        action.data = {
            'deletion': {
                'count': 2,
                'grace_period': 2,
                'candidates': ['NODE_ID_3', 'NODE_ID_4'],
            }
        }
        action.inputs = {}
        mock_count.return_value = 5
        mock_delete.return_value = (action.RES_OK, 'Life is beautiful.')

        # do it
        res_code, res_msg = action.do_scale_in()

        # assertions
        self.assertEqual(action.RES_OK, res_code)
        self.assertEqual('Cluster scaling succeeded.', res_msg)

        # deleting 2 nodes
        mock_count.assert_called_once_with(action.context, 'CID')
        mock_delete.assert_called_once_with(mock.ANY)
        self.assertEqual(2, len(mock_delete.call_args[0][0]))
        self.assertIn('NODE_ID_3', mock_delete.call_args[0][0])
        self.assertIn('NODE_ID_4', mock_delete.call_args[0][0])
        cluster.set_status.assert_called_once_with(
            action.context, consts.CS_RESIZING, 'Cluster scale in started.',
            desired_capacity=3)
        cluster.eval_status.assert_called_once_with(
            action.context, consts.CLUSTER_SCALE_IN)
        mock_sleep.assert_called_once_with(2)

    @mock.patch.object(scaleutils, 'nodes_by_random')
    @mock.patch.object(ca.ClusterAction, '_delete_nodes')
    @mock.patch.object(no.Node, 'count_by_cluster')
    def test_do_scale_in_no_pd_with_input(self, mock_count, mock_delete,
                                          mock_select, mock_load):
        cluster = mock.Mock(id='CID', min_size=1, max_size=-1)
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_ACTION', self.ctx,
                                  data={}, inputs={'count': 3})
        mock_count.return_value = 11
        mock_delete.return_value = (action.RES_OK, 'Life is beautiful.')

        # do it
        res_code, res_msg = action.do_scale_in()

        # assertions
        self.assertEqual(action.RES_OK, res_code)
        self.assertEqual('Cluster scaling succeeded.', res_msg)

        # deleting 3 nodes
        mock_count.assert_called_once_with(action.context, 'CID')
        mock_delete.assert_called_once_with(mock.ANY)
        mock_select.assert_called_once_with(cluster.nodes, 3)
        cluster.set_status.assert_called_once_with(
            action.context, consts.CS_RESIZING, 'Cluster scale in started.',
            desired_capacity=8)
        cluster.eval_status.assert_called_once_with(
            action.context, consts.CLUSTER_SCALE_IN)

    def test_do_scale_in_negative_count(self, mock_load):
        cluster = mock.Mock(id='CID', min_size=1, max_size=-1)
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_ACTION', self.ctx,
                                  data={}, inputs={'count': -3})

        # do it
        res_code, res_msg = action.do_scale_in()

        # assertions
        self.assertEqual(action.RES_ERROR, res_code)
        self.assertEqual('Invalid count (-3) for scaling in.', res_msg)
        self.assertEqual(0, cluster.set_status.call_count)
        self.assertEqual(0, cluster.eval_status.call_count)

    def test_do_scale_in_invalid_count(self, mock_load):
        cluster = mock.Mock(id='CID', min_size=1, max_size=-1)
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_ACTION', self.ctx,
                                  data={}, inputs={'count': 'tt'})

        # do it
        res_code, res_msg = action.do_scale_in()

        # assertions
        self.assertEqual(action.RES_ERROR, res_code)
        self.assertEqual('Invalid count (tt) for scaling in.', res_msg)
        self.assertEqual(0, cluster.set_status.call_count)
        self.assertEqual(0, cluster.eval_status.call_count)

    @mock.patch.object(no.Node, 'count_by_cluster')
    def test_do_scale_in_failed_check(self, mock_count, mock_load):
        cluster = mock.Mock(id='CID', min_size=1, max_size=-1)
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_ACTION', self.ctx,
                                  data={}, inputs={'count': 3})
        mock_count.return_value = 3

        # do it
        res_code, res_msg = action.do_scale_in()

        # assertions
        mock_count.assert_called_once_with(action.context, 'CID')
        self.assertEqual(action.RES_ERROR, res_code)
        self.assertEqual("The target capacity (0) is less than the cluster's "
                         "min_size (1).", res_msg)
        self.assertEqual(0, cluster.set_status.call_count)
        self.assertEqual(0, cluster.eval_status.call_count)

    @mock.patch.object(scaleutils, 'nodes_by_random')
    @mock.patch.object(ca.ClusterAction, '_delete_nodes')
    @mock.patch.object(no.Node, 'count_by_cluster')
    def test_do_scale_in_failed_delete_nodes(self, mock_count, mock_delete,
                                             mock_select, mock_load):

        cluster = mock.Mock(id='CID', min_size=1, max_size=-1)
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_ACTION', self.ctx,
                                  data={}, inputs={'count': 2})
        mock_count.return_value = 5

        # Error cases
        for result in (action.RES_ERROR, action.RES_CANCEL,
                       action.RES_TIMEOUT, action.RES_RETRY):
            mock_delete.return_value = result, 'Too cold to work!'
            # do it
            res_code, res_msg = action.do_scale_in()
            # assertions
            self.assertEqual(result, res_code)
            self.assertEqual('Too cold to work!', res_msg)
            cluster.set_status.assert_called_once_with(
                action.context, consts.CS_RESIZING,
                'Cluster scale in started.',
                desired_capacity=3)
            cluster.eval_status.assert_called_once_with(
                action.context, consts.CLUSTER_SCALE_IN)
            cluster.set_status.reset_mock()
            cluster.eval_status.reset_mock()
            mock_delete.assert_called_once_with(mock.ANY)
            mock_delete.reset_mock()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import random
import string

from oslo_log import log as logging
import six

from senlin.common import context
from senlin.common import exception as exc
from senlin.objects import cluster as cluster_obj
from senlin.profiles.os.nova import server

LOG = logging.getLogger(__name__)


def GenKubeToken():
    token_id = ''.join([random.choice(
        string.digits + string.ascii_lowercase) for i in range(6)])
    token_secret = ''.join([random.choice(
        string.digits + string.ascii_lowercase) for i in range(16)])
    token = '.'.join([token_id, token_secret])
    return token


def loadScript(path):
    script_file = os.path.join(os.path.dirname(__file__), path)
    with open(script_file, "r") as f:
        content = f.read()
    return content


class KubeBaseProfile(server.ServerProfile):
    """Kubernetes Base Profile."""

    def __init__(self, type_name, name, **kwargs):
        super(KubeBaseProfile, self).__init__(type_name, name, **kwargs)
        self.server_id = None

    def _generate_kubeadm_token(self, obj):
        token = GenKubeToken()
        # store generated token

        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        data = obj.data
        data[self.KUBEADM_TOKEN] = token
        cluster_obj.Cluster.update(ctx, obj.id, {'data': data})
        return token

    def _get_kubeadm_token(self, obj):
        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        if obj.cluster_id:
            cluster = cluster_obj.Cluster.get(ctx, obj.cluster_id)
            return cluster.data.get(self.KUBEADM_TOKEN)
        return None

    def _update_master_ip(self, obj, ip):
        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        if obj.cluster_id:
            cluster = cluster_obj.Cluster.get(ctx, obj.cluster_id)
            cluster.data['kube_master_ip'] = ip
            cluster.update(ctx, obj.cluster_id, {'data': cluster.data})

    def _create_network(self, obj):
        client = self.network(obj)
        try:
            net = client.network_create()
            subnet = client.subnet_create(network_id=net.id,
                                          cidr='10.7.0.0/24',
                                          ip_version=4)
        except exc.InternalError as ex:
            raise exc.EResourceCreation(type='kubernetes',
                                        message=six.text_type(ex),
                                        resource_id=obj.id)
        pub_net = client.network_get(self.properties[self.PUBLIC_NETWORK])
        try:
            router = client.router_create(
                external_gateway_info={"network_id": pub_net.id})
            client.add_interface_to_router(router, subnet_id=subnet.id)
            fip = client.floatingip_create(floating_network_id=pub_net.id)
        except exc.InternalError as ex:
            raise exc.EResourceCreation(type='kubernetes',
                                        message=six.text_type(ex),
                                        resource_id=obj.id)

        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        data = obj.data
        data[self.PRIVATE_NETWORK] = net.id
        data[self.PRIVATE_SUBNET] = subnet.id
        data[self.PRIVATE_ROUTER] = router.id
        data[self.KUBE_MASTER_FLOATINGIP] = fip.floating_ip_address
        data[self.KUBE_MASTER_FLOATINGIP_ID] = fip.id

        cluster_obj.Cluster.update(ctx, obj.id, {'data': data})

        return net.id

    def _delete_network(self, obj):
        client = self.network(obj)
        fip_id = obj.data.get(self.KUBE_MASTER_FLOATINGIP_ID)
        if fip_id:
            try:
                # delete floating ip
                client.floatingip_delete(fip_id)
            except exc.InternalError as ex:
                raise exc.EResourceDeletion(type='kubernetes', id=fip_id,
                                            message=six.text_type(ex))

        router = obj.data.get(self.PRIVATE_ROUTER)
        subnet = obj.data.get(self.PRIVATE_SUBNET)
        if router and subnet:
            try:
                client.remove_interface_from_router(router, subnet_id=subnet)
            except exc.InternalError as ex:
                raise exc.EResourceDeletion(type='kubernetes',
                                            id=subnet,
                                            message=six.text_type(ex))

        if router:
            try:
                # delete router
                client.router_delete(router, ignore_missing=True)
            except exc.InternalError as ex:
                raise exc.EResourceDeletion(type='kubernetes',
                                            id=router,
                                            message=six.text_type(ex))

        net = obj.data.get(self.PRIVATE_NETWORK)
        if net:
            try:
                # delete network
                client.network_delete(net, ignore_missing=True)
            except exc.InternalError as ex:
                raise exc.EResourceDeletion(type='kubernetes',
                                            id=net,
                                            message=six.text_type(ex))

    def _associate_floatingip(self, obj, server):
        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)

        if obj.cluster_id:
            cluster = cluster_obj.Cluster.get(ctx, obj.cluster_id)
            fip = cluster.data.get(self.KUBE_MASTER_FLOATINGIP)
            if fip:
                try:
                    self.compute(obj).server_floatingip_associate(server,
                                                                  fip)
                except exc.InternalError as ex:
                    raise exc.EResourceOperation(op='floatingip',
                                                 type='kubernetes',
                                                 id=fip,
                                                 message=six.text_type(ex))

    def _disassociate_floatingip(self, obj, server):
        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        if obj.cluster_id:
            cluster = cluster_obj.Cluster.get(ctx, obj.cluster_id)
            fip = cluster.data.get(self.KUBE_MASTER_FLOATINGIP)
            if fip:
                try:
                    self.compute(obj).server_floatingip_disassociate(server,
                                                                     fip)
                except exc.InternalError as ex:
                    raise exc.EResourceOperation(op='floatingip',
                                                 type='kubernetes',
                                                 id=fip,
                                                 message=six.text_type(ex))

    def _get_cluster_data(self, obj):
        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        if obj.cluster_id:
            cluster = cluster_obj.Cluster.get(ctx, obj.cluster_id)
            return cluster.data
        return {}

    def _get_network(self, obj):
        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        if obj.cluster_id:
            cluster = cluster_obj.Cluster.get(ctx, obj.cluster_id)
            return cluster.data.get(self.PRIVATE_NETWORK)
        return None

    def _create_security_group(self, obj):
        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        sgid = obj.data.get(self.SECURITY_GROUP, None)
        if sgid:
            return sgid

        client = self.network(obj)
        try:
            sg = client.security_group_create(name=self.name)
        except Exception as ex:
            raise exc.EResourceCreation(type='kubernetes',
                                        message=six.text_type(ex))
        data = obj.data
        data[self.SECURITY_GROUP] = sg.id
        cluster_obj.Cluster.update(ctx, obj.id, {'data': data})
        self._set_security_group_rules(obj, sg.id)

        return sg.id

    def _get_security_group(self, obj):
        ctx = context.get_service_context(user_id=obj.user,
                                          project_id=obj.project)
        if obj.cluster_id:
            cluster = cluster_obj.Cluster.get(ctx, obj.cluster_id)
            return cluster.data.get(self.SECURITY_GROUP)
        return None

    def _set_security_group_rules(self, obj, sgid):
        client = self.network(obj)
        open_ports = {
            'tcp': [22, 80, 8000, 8080, 6443, 8001, 8443, 443,
                    179, 8082, 8086],
            'udp': [8285, 8472],
            'icmp': [None]
        }
        for p in open_ports.keys():
            for port in open_ports[p]:
                try:
                    client.security_group_rule_create(sgid, port, protocol=p)
                except Exception as ex:
                    raise exc.EResourceCreation(type='kubernetes',
                                                message=six.text_type(ex))

    def _delete_security_group(self, obj):
        sgid = obj.data.get(self.SECURITY_GROUP)
        if sgid:
            try:
                self.network(obj).security_group_delete(sgid,
                                                        ignore_missing=True)
            except exc.InternalError as ex:
                raise exc.EResourceDeletion(type='kubernetes',
                                            id=sgid,
                                            message=six.text_type(ex))

    def do_validate(self, obj):
        """Validate if the spec has provided valid info for server creation.

        :param obj: The node object.
        """
        # validate flavor
        flavor = self.properties[self.FLAVOR]
        self._validate_flavor(obj, flavor)

        # validate image
        image = self.properties[self.IMAGE]
        if image is not None:
            self._validate_image(obj, image)

        # validate key_name
        keypair = self.properties[self.KEY_NAME]
        if keypair is not None:
            self._validate_keypair(obj, keypair)

        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import copy

from oslo_config import cfg

from senlin.objects.requests import nodes
from senlin.tests.unit.common import base as test_base

CONF = cfg.CONF
CONF.import_opt('default_action_timeout', 'senlin.common.config')


class TestNodeCreate(test_base.SenlinTestCase):

    body = {
        'name': 'test-node',
        'profile_id': 'test-profile',
    }

    def test_node_create_request_body(self):
        sot = nodes.NodeCreateRequestBody(**self.body)
        self.assertEqual('test-node', sot.name)
        self.assertEqual('test-profile', sot.profile_id)

        sot.obj_set_defaults()

        self.assertEqual('', sot.cluster_id)
        self.assertEqual('', sot.role)
        self.assertEqual({}, sot.metadata)

    def test_node_create_request_body_full(self):
        body = copy.deepcopy(self.body)
        body['role'] = 'master'
        body['cluster_id'] = 'cluster-01'
        body['metadata'] = {'foo': 'bar'}
        sot = nodes.NodeCreateRequestBody(**body)
        self.assertEqual('test-node', sot.name)
        self.assertEqual('test-profile', sot.profile_id)
        self.assertEqual('cluster-01', sot.cluster_id)
        self.assertEqual('master', sot.role)
        self.assertEqual({'foo': 'bar'}, sot.metadata)

    def test_request_body_to_primitive(self):
        sot = nodes.NodeCreateRequestBody(**self.body)
        res = sot.obj_to_primitive()
        self.assertEqual(
            {
                'name': u'test-node',
                'profile_id': u'test-profile'
            },
            res['senlin_object.data']
        )
        self.assertEqual('NodeCreateRequestBody',
                         res['senlin_object.name'])
        self.assertEqual('senlin', res['senlin_object.namespace'])
        self.assertEqual('1.0', res['senlin_object.version'])
        self.assertIn('profile_id', res['senlin_object.changes'])
        self.assertIn('name', res['senlin_object.changes'])

    def test_request_to_primitive(self):
        body = nodes.NodeCreateRequestBody(**self.body)
        request = {'node': body}
        sot = nodes.NodeCreateRequest(**request)
        self.assertIsInstance(sot.node, nodes.NodeCreateRequestBody)

        self.assertEqual('test-node', sot.node.name)
        self.assertEqual('test-profile', sot.node.profile_id)

        res = sot.obj_to_primitive()
        self.assertEqual(['node'], res['senlin_object.changes'])
        self.assertEqual('NodeCreateRequest', res['senlin_object.name'])
        self.assertEqual('senlin', res['senlin_object.namespace'])
        self.assertEqual('1.0', res['senlin_object.version'])
        data = res['senlin_object.data']['node']
        self.assertIn('profile_id', data['senlin_object.changes'])
        self.assertIn('name', data['senlin_object.changes'])
        self.assertEqual('NodeCreateRequestBody',
                         data['senlin_object.name'])
        self.assertEqual('senlin', data['senlin_object.namespace'])
        self.assertEqual('1.0', data['senlin_object.version'])
        self.assertEqual(
            {'name': u'test-node', 'profile_id': u'test-profile'},
            data['senlin_object.data']
        )


class TestNodeList(test_base.SenlinTestCase):

    def test_node_list_request_body_full(self):
        params = {
            'cluster_id': '8c3c9af7-d768-4c5a-a21e-5261b22d749d',
            'name': ['node01'],
            'status': ['ACTIVE'],
            'limit': 3,
            'marker': 'f1ed0d50-7651-4599-a8cb-c86e9c7123f5',
            'sort': 'name:asc',
            'project_safe': False,
        }
        sot = nodes.NodeListRequest(**params)
        self.assertEqual('8c3c9af7-d768-4c5a-a21e-5261b22d749d',
                         sot.cluster_id)
        self.assertEqual(['node01'], sot.name)
        self.assertEqual(['ACTIVE'], sot.status)
        self.assertEqual(3, sot.limit)
        self.assertEqual('f1ed0d50-7651-4599-a8cb-c86e9c7123f5', sot.marker)
        self.assertEqual('name:asc', sot.sort)
        self.assertFalse(sot.project_safe)

    def test_node_list_request_body_default(self):
        sot = nodes.NodeListRequest()
        sot.obj_set_defaults()
        self.assertTrue(sot.project_safe)


class TestNodeGet(test_base.SenlinTestCase):

    def test_node_get_request_full(self):
        params = {
            'identity': 'node-001',
            'show_details': True,
        }
        sot = nodes.NodeGetRequest(**params)
        self.assertEqual('node-001', sot.identity)
        self.assertTrue(sot.show_details)

    def test_node_get_request_default(self):
        sot = nodes.NodeGetRequest()
        sot.obj_set_defaults()
        self.assertFalse(sot.show_details)


class TestNodeUpdate(test_base.SenlinTestCase):

    body = {
        'identity': 'test-node',
        'name': 'test-node-newname',
        'profile_id': 'test-profile',
        'metadata': {'foo': 'bar'},
        'role': 'master'
    }

    def test_node_update_request(self):
        sot = nodes.NodeUpdateRequest(**self.body)
        self.assertEqual('test-node', sot.identity)
        self.assertEqual('test-node-newname', sot.name)
        self.assertEqual('test-profile', sot.profile_id)
        self.assertEqual('master', sot.role)
        self.assertEqual({'foo': 'bar'}, sot.metadata)


class TestNodeDelete(test_base.SenlinTestCase):

    body = {
        'identity': 'test-node'
    }

    def test_node_delete_request(self):
        sot = nodes.NodeDeleteRequest(**self.body)
        self.assertEqual('test-node', sot.identity)


class TestNodeCheck(test_base.SenlinTestCase):

    body = {
        'identity': 'test-node',
        'params': {'foo': 'bar'},
    }

    def test_node_check_request(self):
        sot = nodes.NodeCheckRequest(**self.body)
        self.assertEqual({'foo': 'bar'}, sot.params)


class TestNodeRecover(test_base.SenlinTestCase):

    body = {
        'identity': 'test-node',
        'params': {'foo': 'bar'},
    }

    def test_node_recover_request(self):
        sot = nodes.NodeRecoverRequest(**self.body)
        self.assertEqual({'foo': 'bar'}, sot.params)


class TestNodeOperation(test_base.SenlinTestCase):

    body = {
        'identity': 'test-node',
        'operation': 'dance',
        'params': {'foo': 'bar'},
    }

    def test_node_operation_request(self):
        sot = nodes.NodeOperationRequest(**self.body)
        self.assertEqual('test-node', sot.identity)
        self.assertEqual('dance', sot.operation)
        self.assertEqual({'foo': 'bar'}, sot.params)


class TestNodeAdopt(test_base.SenlinTestCase):

    body = {
        'identity': 'test-node',
        'type': 'test-type',
        'name': 'test-name',
        'cluster': 'test-cluster',
        'role': 'test-role',
        'metadata': {'key': 'value'},
        'overrides': {'foo': 'bar'},
        'snapshot': True
    }

    def test_node_adopt_request(self):
        sot = nodes.NodeAdoptRequest(**self.body)
        self.assertEqual('test-node', sot.identity)
        self.assertEqual('test-type', sot.type)
        self.assertEqual('test-name', sot.name)
        self.assertEqual('test-cluster', sot.cluster)
        self.assertEqual('test-role', sot.role)
        self.assertEqual({'key': 'value'}, sot.metadata)
        self.assertEqual({'foo': 'bar'}, sot.overrides)
        self.assertTrue(sot.snapshot)


class TestNodeAdoptPreview(test_base.SenlinTestCase):

    body = {
        'identity': 'test-node',
        'type': 'test-type',
        'overrides': {'foo': 'bar'},
        'snapshot': True
    }

    def test_node_adopt_request(self):
        sot = nodes.NodeAdoptPreviewRequest(**self.body)
        self.assertEqual('test-node', sot.identity)
        self.assertEqual('test-type', sot.type)
        self.assertEqual({'foo': 'bar'}, sot.overrides)
        self.assertTrue(sot.snapshot)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlin.drivers import base
from senlin.drivers import sdk


class OctaviaClient(base.DriverBase):
    """Octavia v2 client"""

    def __init__(self, params):
        super(OctaviaClient, self).__init__(params)
        self.conn = sdk.create_connection(params)

    @sdk.translate_exception
    def loadbalancer_get(self, name_or_id, ignore_missing=False,
                         show_deleted=False):
        lb = self.conn.load_balancer.find_load_balancer(name_or_id,
                                                        ignore_missing)
        # TODO(liyi)
        # It's unreasonable for octavia don't support filter deleted
        # loadbalancers. So if supported, we need to change the function.
        if lb and not show_deleted and lb.provisioning_status == 'DELETED':
            lb = None

        return lb

    @sdk.translate_exception
    def loadbalancer_create(self, vip_subnet_id, vip_address=None,
                            admin_state_up=True, name=None, description=None):

        kwargs = {
            'vip_subnet_id': vip_subnet_id,
            'admin_state_up': admin_state_up,
        }

        if vip_address is not None:
            kwargs['vip_address'] = vip_address
        if name is not None:
            kwargs['name'] = name
        if description is not None:
            kwargs['description'] = description

        res = self.conn.load_balancer.create_load_balancer(**kwargs)
        return res

    @sdk.translate_exception
    def loadbalancer_delete(self, lb_id, ignore_missing=True):
        self.conn.load_balancer.delete_load_balancer(
            lb_id, ignore_missing=ignore_missing)
        return

    @sdk.translate_exception
    def listener_create(self, loadbalancer_id, protocol, protocol_port,
                        connection_limit=None,
                        admin_state_up=True, name=None, description=None):

        kwargs = {
            'loadbalancer_id': loadbalancer_id,
            'protocol': protocol,
            'protocol_port': protocol_port,
            'admin_state_up': admin_state_up,
        }

        if connection_limit is not None:
            kwargs['connection_limit'] = connection_limit
        if name is not None:
            kwargs['name'] = name
        if description is not None:
            kwargs['description'] = description

        res = self.conn.load_balancer.create_listener(**kwargs)
        return res

    @sdk.translate_exception
    def listener_delete(self, listener_id, ignore_missing=True):
        self.conn.load_balancer.delete_listener(
            listener_id, ignore_missing=ignore_missing)
        return

    @sdk.translate_exception
    def pool_create(self, lb_algorithm, listener_id, protocol,
                    admin_state_up=True, name=None, description=None):

        kwargs = {
            'lb_algorithm': lb_algorithm,
            'listener_id': listener_id,
            'protocol': protocol,
            'admin_state_up': admin_state_up,
        }

        if name is not None:
            kwargs['name'] = name
        if description is not None:
            kwargs['description'] = description

        res = self.conn.load_balancer.create_pool(**kwargs)
        return res

    @sdk.translate_exception
    def pool_delete(self, pool_id, ignore_missing=True):
        self.conn.load_balancer.delete_pool(
            pool_id, ignore_missing=ignore_missing)
        return

    @sdk.translate_exception
    def pool_member_create(self, pool_id, address, protocol_port, subnet_id,
                           weight=None, admin_state_up=True):

        kwargs = {
            'address': address,
            'protocol_port': protocol_port,
            'admin_state_up': admin_state_up,
            'subnet_id': subnet_id,
        }

        if weight is not None:
            kwargs['weight'] = weight

        res = self.conn.load_balancer.create_member(pool_id, **kwargs)
        return res

    @sdk.translate_exception
    def pool_member_delete(self, pool_id, member_id, ignore_missing=True):
        self.conn.load_balancer.delete_member(
            member_id, pool_id, ignore_missing=ignore_missing)
        return

    @sdk.translate_exception
    def healthmonitor_create(self, hm_type, delay, timeout, max_retries,
                             pool_id, admin_state_up=True,
                             http_method=None, url_path=None,
                             expected_codes=None):
        kwargs = {
            'type': hm_type,
            'delay': delay,
            'timeout': timeout,
            'max_retries': max_retries,
            'pool_id': pool_id,
            'admin_state_up': admin_state_up,
        }

        # TODO(anyone): verify if this is correct
        if hm_type == 'HTTP':
            if http_method is not None:
                kwargs['http_method'] = http_method
            if url_path is not None:
                kwargs['url_path'] = url_path
            if expected_codes is not None:
                kwargs['expected_codes'] = expected_codes

        res = self.conn.load_balancer.create_health_monitor(**kwargs)
        return res

    @sdk.translate_exception
    def healthmonitor_delete(self, hm_id, ignore_missing=True):
        self.conn.load_balancer.delete_health_monitor(
            hm_id, ignore_missing=ignore_missing)
        return



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlin.drivers import base


class LoadBalancerDriver(base.DriverBase):
    def __init__(self, params):
        self.lb_result = {
            "loadbalancer": "a36c20d0-18e9-42ce-88fd-82a35977ee8c",
            "vip_address": "192.168.1.100",
            "listener": "35cb8516-1173-4035-8dae-0dae3453f37f",
            "pool": "4c0a0a5f-cf8f-44b7-b912-957daa8ce5e5",
            "healthmonitor": "0a9ac99d-0a09-4b18-8499-a0796850279a"
        }

        self.member_id = "9a7aff27-fd41-4ec1-ba4c-3eb92c629313"

    def lb_create(self, vip, pool, hm=None):
        return True, self.lb_result

    def lb_delete(self, **kwargs):
        return True, 'LB deletion succeeded'

    def member_add(self, node, lb_id, pool_id, port, subnet):
        return self.member_id

    def member_remove(self, lb_id, pool_id, member_id):
        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
import six

from senlin.common import consts
from senlin.common import exception
from senlin.engine import node as nodem
from senlin.objects import node as node_obj
from senlin.profiles import base as pb
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils

PROFILE_ID = 'aa5f86b8-e52b-4f2b-828a-4c14c770938d'
CLUSTER_ID = '2c5139a6-24ba-4a6f-bd53-a268f61536de'
NODE_ID = '60efdaa1-06c2-4fcf-ae44-17a2d85ff3ea'


class TestNode(base.SenlinTestCase):

    def setUp(self):
        super(TestNode, self).setUp()
        self.context = utils.dummy_context(project='node_test_project')
        self.profile = utils.create_profile(self.context, PROFILE_ID)
        self.cluster = utils.create_cluster(self.context, CLUSTER_ID,
                                            PROFILE_ID)

    def test_node_init(self):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, role='first_node')
        self.assertIsNone(node.id)
        self.assertEqual('node1', node.name)
        self.assertIsNone(node.physical_id)
        self.assertEqual(PROFILE_ID, node.profile_id)
        self.assertEqual('', node.user)
        self.assertEqual('', node.project)
        self.assertEqual('', node.domain)
        self.assertEqual(CLUSTER_ID, node.cluster_id)
        self.assertEqual(-1, node.index)
        self.assertEqual('first_node', node.role)

        self.assertIsNone(node.init_at)
        self.assertIsNone(node.created_at)
        self.assertIsNone(node.updated_at)

        self.assertEqual('INIT', node.status)
        self.assertEqual('Initializing', node.status_reason)
        self.assertEqual({}, node.data)
        self.assertEqual({}, node.metadata)
        self.assertEqual({}, node.rt)

    def test_node_init_random_name(self):
        node = nodem.Node(None, PROFILE_ID, None)
        self.assertIsNotNone(node.name)
        self.assertEqual(13, len(node.name))

    def test_node_store_init(self):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context,
                          role='first_node', index=1)
        self.assertIsNone(node.id)
        node_id = node.store(self.context)
        self.assertIsNotNone(node_id)

        node_info = node_obj.Node.get(self.context, node_id)
        self.assertIsNotNone(node_info)
        self.assertEqual('node1', node_info.name)
        self.assertIsNone(node_info.physical_id)
        self.assertEqual(CLUSTER_ID, node_info.cluster_id)
        self.assertEqual(PROFILE_ID, node_info.profile_id)
        self.assertEqual(self.context.user_id, node_info.user)
        self.assertEqual(self.context.project_id, node_info.project)
        self.assertEqual(self.context.domain_id, node_info.domain)
        self.assertEqual(1, node_info.index)
        self.assertEqual('first_node', node.role)

        self.assertIsNotNone(node_info.init_at)
        self.assertIsNone(node_info.created_at)
        self.assertIsNone(node_info.updated_at)

        self.assertEqual('INIT', node_info.status)
        self.assertEqual('Initializing', node_info.status_reason)
        self.assertEqual({}, node_info.metadata)
        self.assertEqual({}, node_info.data)

    def test_node_store_update(self):
        node = nodem.Node('node1', PROFILE_ID, "", user=self.context.user_id,
                          project=self.context.project_id)
        node_id = node.store(self.context)

        node.name = 'new_name'
        new_node_id = node.store(self.context)

        self.assertEqual(node_id, new_node_id)

    def test_node_load(self):
        ex = self.assertRaises(exception.ResourceNotFound,
                               nodem.Node.load,
                               self.context, 'non-existent', None)
        self.assertEqual("The node 'non-existent' could not be found.",
                         six.text_type(ex))

        x_node_id = 'ee96c490-2dee-40c8-8919-4c64b89e326c'
        node = utils.create_node(self.context, x_node_id, PROFILE_ID,
                                 CLUSTER_ID)
        node_info = nodem.Node.load(self.context, x_node_id)

        self.assertEqual(node.id, node_info.id)
        self.assertEqual(node.name, node_info.name)
        self.assertEqual(node.physical_id, node_info.physical_id)
        self.assertEqual(node.cluster_id, node_info.cluster_id)
        self.assertEqual(node.profile_id, node_info.profile_id)
        self.assertEqual(node.user, node_info.user)
        self.assertEqual(node.project, node_info.project)
        self.assertEqual(node.domain, node_info.domain)
        self.assertEqual(node.index, node_info.index)
        self.assertEqual(node.role, node_info.role)

        self.assertEqual(node.init_at, node_info.init_at)
        self.assertEqual(node.created_at, node_info.created_at)
        self.assertEqual(node.updated_at, node_info.updated_at)

        self.assertEqual(node.status, node_info.status)
        self.assertEqual(node.status_reason, node_info.status_reason)
        self.assertEqual(node.metadata, node_info.metadata)
        self.assertEqual(node.data, node_info.data)
        self.assertEqual(self.profile.name, node_info.rt['profile'].name)

    def test_node_load_diff_project(self):
        x_node_id = 'c06840c5-f4e4-49ae-8143-9da5b4c73f38'
        utils.create_node(self.context, x_node_id, PROFILE_ID, CLUSTER_ID)

        new_ctx = utils.dummy_context(project='a-different-project')
        ex = self.assertRaises(exception.ResourceNotFound,
                               nodem.Node.load,
                               new_ctx, x_node_id, None)
        self.assertEqual("The node '%s' could not be found." % x_node_id,
                         six.text_type(ex))

        res = nodem.Node.load(new_ctx, x_node_id, project_safe=False)
        self.assertIsNotNone(res)
        self.assertEqual(x_node_id, res.id)

    @mock.patch.object(nodem.Node, '_from_object')
    @mock.patch.object(node_obj.Node, 'get_all')
    def test_node_load_all(self, mock_get, mock_init):
        x_obj_1 = mock.Mock()
        x_obj_2 = mock.Mock()
        mock_get.return_value = [x_obj_1, x_obj_2]

        x_node_1 = mock.Mock()
        x_node_2 = mock.Mock()
        mock_init.side_effect = [x_node_1, x_node_2]

        result = nodem.Node.load_all(self.context)

        self.assertEqual([x_node_1, x_node_2], [n for n in result])
        mock_get.assert_called_once_with(self.context, cluster_id=None,
                                         limit=None, marker=None,
                                         sort=None, filters=None,
                                         project_safe=True)
        mock_init.assert_has_calls([
            mock.call(self.context, x_obj_1),
            mock.call(self.context, x_obj_2)])

    def test_node_set_status(self):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        node.store(self.context)
        self.assertEqual(nodem.consts.NS_INIT, node.status)
        self.assertIsNotNone(node.init_at)
        self.assertIsNone(node.created_at)
        self.assertIsNone(node.updated_at)

        # create
        node.set_status(self.context, consts.NS_CREATING,
                        reason='Creation in progress')
        self.assertEqual('CREATING', node.status)
        self.assertEqual('Creation in progress', node.status_reason)
        self.assertIsNone(node.created_at)
        self.assertIsNone(node.updated_at)

        node.set_status(self.context, consts.NS_ACTIVE,
                        reason='Creation succeeded')
        self.assertEqual('ACTIVE', node.status)
        self.assertEqual('Creation succeeded', node.status_reason)
        self.assertIsNotNone(node.created_at)
        self.assertIsNotNone(node.updated_at)

        # update
        node.set_status(self.context, consts.NS_UPDATING,
                        reason='Update in progress')
        self.assertEqual('UPDATING', node.status)
        self.assertEqual('Update in progress', node.status_reason)
        self.assertIsNotNone(node.created_at)

        node.set_status(self.context, consts.NS_ACTIVE,
                        reason='Update succeeded')
        self.assertEqual('ACTIVE', node.status)
        self.assertEqual('Update succeeded', node.status_reason)
        self.assertIsNotNone(node.created_at)
        self.assertIsNotNone(node.updated_at)

        node.set_status(self.context, consts.NS_ACTIVE)
        self.assertEqual('ACTIVE', node.status)
        self.assertIsNotNone(node.created_at)
        self.assertIsNotNone(node.updated_at)

        # delete
        node.set_status(self.context, consts.NS_DELETING,
                        reason='Deletion in progress')
        self.assertEqual('DELETING', node.status)
        self.assertEqual('Deletion in progress', node.status_reason)
        self.assertIsNotNone(node.created_at)

    @mock.patch.object(pb.Profile, 'get_details')
    def test_node_get_details(self, mock_details):
        node = nodem.Node('node1', CLUSTER_ID, None)
        for physical_id in (None, ''):
            node.physical_id = physical_id
            self.assertEqual({}, node.get_details(self.context))
            self.assertEqual(0, mock_details.call_count)

        fake_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        node.physical_id = fake_id
        mock_details.return_value = {'foo': 'bar'}
        res = node.get_details(self.context)
        mock_details.assert_called_once_with(self.context, node)
        self.assertEqual({'foo': 'bar'}, res)

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'create_object')
    def test_node_create(self, mock_create, mock_status):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        mock_create.return_value = physical_id
        res = node.do_create(self.context)
        self.assertTrue(res)
        mock_status.assert_any_call(self.context, consts.NS_CREATING,
                                    'Creation in progress')
        mock_status.assert_any_call(self.context, consts.NS_ACTIVE,
                                    'Creation succeeded',
                                    physical_id=physical_id)

    def test_node_create_not_init(self):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        node.status = 'NOT_INIT'
        res = node.do_create(self.context)
        self.assertFalse(res)

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'create_object')
    def test_node_create_not_created(self, mock_create, mock_status):

        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        mock_create.side_effect = exception.EResourceCreation(
            type='PROFILE', message='Boom', resource_id='test_id')

        res = node.do_create(self.context)

        self.assertFalse(res)
        mock_status.assert_any_call(self.context, consts.NS_CREATING,
                                    'Creation in progress')
        mock_status.assert_any_call(self.context, consts.NS_ERROR,
                                    'Failed in creating PROFILE: Boom.',
                                    physical_id='test_id')

    @mock.patch.object(node_obj.Node, 'delete')
    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'delete_object')
    def test_node_delete(self, mock_delete, mock_status, mock_db_delete):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        node.physical_id = uuidutils.generate_uuid()
        node.id = uuidutils.generate_uuid()
        mock_db_delete.return_value = True

        res = node.do_delete(self.context)

        self.assertTrue(res)
        mock_delete.assert_called_once_with(self.context, node)
        mock_db_delete.assert_called_once_with(self.context, node.id)
        mock_status.assert_called_once_with(self.context, consts.NS_DELETING,
                                            'Deletion in progress')

    @mock.patch.object(node_obj.Node, 'delete')
    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'delete_object')
    def test_node_delete_no_physical_id(self, mock_delete, mock_status,
                                        mock_db_delete):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        node.id = uuidutils.generate_uuid()
        self.assertIsNone(node.physical_id)
        mock_db_delete.return_value = True

        res = node.do_delete(self.context)

        self.assertTrue(res)
        mock_status.assert_called_once_with(self.context, consts.NS_DELETING,
                                            "Deletion in progress")
        self.assertTrue(mock_delete.called)
        mock_db_delete.assert_called_once_with(self.context, node.id)

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'delete_object')
    def test_node_delete_EResourceDeletion(self, mock_delete, mock_status):
        ex = exception.EResourceDeletion(type='PROFILE', id='NODE_ID',
                                         message='Too Bad')
        mock_delete.side_effect = ex
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        node.physical_id = uuidutils.generate_uuid()

        res = node.do_delete(self.context)

        self.assertFalse(res)
        mock_delete.assert_called_once_with(self.context, node)
        mock_status.assert_has_calls([
            mock.call(self.context, consts.NS_DELETING,
                      "Deletion in progress"),
            mock.call(self.context, consts.NS_ERROR,
                      "Failed in deleting PROFILE 'NODE_ID': Too Bad.")
        ])

    @mock.patch.object(node_obj.Node, 'update')
    @mock.patch.object(pb.Profile, 'update_object')
    def test_node_update_new_profile(self, mock_update, mock_db):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context,
                          physical_id=uuidutils.generate_uuid())
        node.id = node.store(self.context)
        new_id = uuidutils.generate_uuid()
        utils.create_profile(self.context, new_id)
        mock_update.return_value = True

        res = node.do_update(self.context, {'new_profile_id': new_id})

        self.assertTrue(res)
        mock_update.assert_called_once_with(self.context, node, new_id)
        self.assertEqual(new_id, node.profile_id)
        self.assertEqual(new_id, node.rt['profile'].id)
        mock_db.assert_has_calls([
            mock.call(self.context, node.id,
                      {'status': consts.NS_UPDATING,
                       'status_reason': 'Update in progress'}),
            mock.call(self.context, node.id,
                      {'status': consts.NS_ACTIVE,
                       'status_reason': 'Update succeeded',
                       'profile_id': new_id,
                       'updated_at': mock.ANY})
        ])

    @mock.patch.object(pb.Profile, 'update_object')
    @mock.patch.object(node_obj.Node, 'update')
    def test_node_update_name(self, mock_db, mock_update):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        node.store(self.context)

        physical_id = uuidutils.generate_uuid()
        node.physical_id = physical_id

        res = node.do_update(self.context, {'name': 'new_name',
                                            'role': 'new_role',
                                            'metadata': {'k': {'m': 'v'}},
                                            'bogus': 'foo'})

        self.assertTrue(res)
        self.assertEqual('new_name', node.name)
        mock_db.assert_has_calls([
            mock.call(self.context, node.id,
                      {'status': 'UPDATING',
                       'status_reason': 'Update in progress'}),
            mock.call(self.context, node.id,
                      {'status': 'ACTIVE',
                       'status_reason': 'Update succeeded',
                       'name': 'new_name',
                       'role': 'new_role',
                       'metadata': {'k': {'m': 'v'}},
                       'updated_at': mock.ANY})
        ])
        self.assertEqual(0, mock_update.call_count)

    def test_node_update_not_created(self):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        self.assertIsNone(node.physical_id)
        new_profile_id = '71d8f4dd-1ef9-4308-b7ae-03298b04449e'
        res = node.do_update(self.context, new_profile_id)
        self.assertFalse(res)

    @mock.patch.object(pb.Profile, 'update_object')
    @mock.patch.object(node_obj.Node, 'update')
    def test_node_update_EResourceUpdate(self, mock_db, mock_update):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        node.physical_id = uuidutils.generate_uuid()
        node.id = uuidutils.generate_uuid()

        ex = exception.EResourceUpdate(type='PROFILE', id='ID',
                                       message='reason')
        mock_update.side_effect = ex
        new_id = uuidutils.generate_uuid()
        utils.create_profile(self.context, new_id)

        res = node.do_update(self.context, {'new_profile_id': new_id})

        self.assertFalse(res)
        self.assertNotEqual(new_id, node.profile_id)
        mock_db.assert_has_calls([
            mock.call(
                self.context, node.id,
                {"status": "UPDATING", "status_reason": "Update in progress"}
            ),
            mock.call(
                self.context, node.id,
                {"status": "ERROR",
                 "status_reason": "Failed in updating PROFILE 'ID': reason.",
                 "updated_at": mock.ANY}
            )
        ])
        self.assertEqual(1, mock_update.call_count)

    @mock.patch.object(node_obj.Node, 'migrate')
    def test_node_join_same_cluster(self, mock_migrate):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        node.index = 1
        res = node.do_join(self.context, CLUSTER_ID)
        self.assertTrue(res)
        self.assertEqual(1, node.index)
        self.assertIsNone(node.updated_at)
        self.assertFalse(mock_migrate.called)

    @mock.patch.object(pb.Profile, 'join_cluster')
    @mock.patch.object(node_obj.Node, 'migrate')
    def test_node_join(self, mock_migrate, mock_join_cluster):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        mock_join_cluster.return_value = True
        cluster_id = 'fb8bca7a-a82b-4442-a40f-92d3e3cfb0b9'

        res = node.do_join(self.context, cluster_id)

        self.assertTrue(res)
        mock_migrate.assert_called_once_with(self.context, node.id,
                                             cluster_id, mock.ANY)
        mock_join_cluster.assert_called_once_with(self.context, node,
                                                  cluster_id)
        self.assertEqual(cluster_id, node.cluster_id)
        self.assertEqual(mock_migrate.return_value.index, node.index)
        self.assertIsNotNone(node.updated_at)

    @mock.patch.object(pb.Profile, 'join_cluster')
    def test_node_join_fail_profile_call(self, mock_join):
        node = nodem.Node('node1', PROFILE_ID, None, self.context)
        node.id = uuidutils.generate_uuid()
        mock_join.return_value = False
        cluster_id = 'fb8bca7a-a82b-4442-a40f-92d3e3cfb0b9'

        res = node.do_join(self.context, cluster_id)

        self.assertFalse(res)
        mock_join.assert_called_once_with(self.context, node, cluster_id)
        self.assertEqual(-1, node.index)

    @mock.patch.object(node_obj.Node, 'migrate')
    def test_node_leave_no_cluster(self, mock_migrate):
        node = nodem.Node('node1', PROFILE_ID, '', self.context)
        self.assertTrue(node.do_leave(self.context))
        self.assertFalse(mock_migrate.called)
        self.assertEqual('', node.cluster_id)
        self.assertIsNone(node.updated_at)

    @mock.patch.object(pb.Profile, 'leave_cluster')
    @mock.patch.object(node_obj.Node, 'migrate')
    def test_node_leave(self, mock_migrate, mock_leave_cluster):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context)
        mock_leave_cluster.return_value = True
        res = node.do_leave(self.context)
        self.assertTrue(res)
        self.assertEqual('', node.cluster_id)
        self.assertIsNotNone(node.updated_at)
        self.assertEqual(-1, node.index)
        mock_migrate.assert_called_once_with(self.context, node.id,
                                             None, mock.ANY)
        mock_leave_cluster.assert_called_once_with(self.context, node)

    @mock.patch.object(pb.Profile, 'leave_cluster')
    def test_node_leave_fail_update_server_metadata(self, mock_leave):
        node = nodem.Node('node1', PROFILE_ID, CLUSTER_ID, self.context,
                          index=1)
        mock_leave.return_value = False
        res = node.do_leave(self.context)
        self.assertFalse(res)
        self.assertNotEqual('', node.cluster_id)
        self.assertIsNone(node.updated_at)
        self.assertEqual(1, node.index)

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'check_object')
    def test_node_check(self, mock_check, mock_status):
        node = nodem.Node('node1', PROFILE_ID, '')
        node.status = consts.NS_ACTIVE
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        mock_check.return_value = True
        res = node.do_check(self.context)

        self.assertTrue(res)
        mock_check.assert_called_once_with(self.context, node)
        mock_status.assert_called_once_with(self.context, consts.NS_ACTIVE,
                                            'Check: Node is ACTIVE.')

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'check_object')
    def test_node_check_warning(self, mock_check, mock_status):
        node = nodem.Node('node1', PROFILE_ID, '')
        node.status = consts.NS_WARNING
        node.status_reason = 'bad news'
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        mock_check.return_value = True

        res = node.do_check(self.context)

        self.assertTrue(res)
        mock_check.assert_called_once_with(self.context, node)
        msg = ("Check: Physical object is ACTIVE but the node status was "
               "WARNING. %s") % node.status_reason
        mock_status.assert_called_once_with(self.context, consts.NS_WARNING,
                                            msg)

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'check_object')
    def test_node_check_not_active(self, mock_check, mock_status):
        node = nodem.Node('node1', PROFILE_ID, '')
        node.status = consts.NS_WARNING
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        mock_check.return_value = False

        res = node.do_check(self.context)

        self.assertTrue(res)
        mock_status.assert_called_once_with(self.context, consts.NS_ERROR,
                                            'Check: Node is not ACTIVE.')

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'check_object')
    def test_node_check_check_with_exc(self, mock_check, mock_status):
        node = nodem.Node('node1', PROFILE_ID, '')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        err = exception.EResourceOperation(op='checking', type='server',
                                           id=node.physical_id,
                                           message='failed get')
        mock_check.side_effect = err

        res = node.do_check(self.context)

        self.assertFalse(res)
        mock_status.assert_called_once_with(
            self.context,
            consts.NS_ERROR,
            "Failed in checking server '%s': failed get." % node.physical_id)

    def test_node_check_no_physical_id(self):
        node = nodem.Node('node1', PROFILE_ID, '')

        res = node.do_check(self.context)

        self.assertFalse(res)

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'check_object')
    def test_node_check_no_server(self, mock_check, mock_status):
        node = nodem.Node('node1', PROFILE_ID, '')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'

        err = exception.EServerNotFound(type='server',
                                        id=node.physical_id,
                                        message='No Server found')
        mock_check.side_effect = err

        res = node.do_check(self.context)

        self.assertTrue(res)
        mock_status.assert_called_once_with(
            self.context, consts.NS_ERROR,
            "Failed in found server '%s': No Server found."
            % node.physical_id,
            physical_id=None)

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'recover_object')
    def test_node_recover_new_object(self, mock_recover, mock_status):
        def set_status(*args, **kwargs):
            if args[1] == 'ACTIVE':
                node.physical_id = new_id
                node.data = {'recovery': 'RECREATE'}

        node = nodem.Node('node1', PROFILE_ID, '')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        new_id = '166db83b-b4a4-49ef-96a8-6c0fdd882d1a'

        # action = node_action.NodeAction(node.id, 'ACTION', self.ctx)

        mock_recover.return_value = new_id, True
        mock_status.side_effect = set_status
        action = mock.Mock()
        action.inputs = {'operation': ['SWIM', 'DANCE']}

        res = node.do_recover(self.context, action)

        self.assertTrue(res)
        mock_recover.assert_called_once_with(
            self.context, node, **action.inputs)
        self.assertEqual('node1', node.name)
        self.assertEqual(new_id, node.physical_id)
        self.assertEqual(PROFILE_ID, node.profile_id)
        self.assertEqual({'recovery': 'RECREATE'}, node.data)
        mock_status.assert_has_calls([
            mock.call(self.context, 'RECOVERING',
                      reason='Recovery in progress'),
            mock.call(self.context, consts.NS_ACTIVE,
                      reason='Recovery succeeded',
                      physical_id=new_id,
                      data={'recovery': 'RECREATE'})])

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'recover_object')
    def test_node_recover_in_place(self, mock_recover, mock_status):
        node = nodem.Node('node1', PROFILE_ID, None)
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        mock_recover.return_value = node.physical_id, True
        action = mock.Mock(inputs={})

        res = node.do_recover(self.context, action)

        self.assertTrue(res)
        mock_recover.assert_called_once_with(self.context, node)
        self.assertEqual('node1', node.name)
        self.assertEqual(PROFILE_ID, node.profile_id)
        mock_status.assert_has_calls([
            mock.call(self.context, 'RECOVERING',
                      reason='Recovery in progress'),
            mock.call(self.context, consts.NS_ACTIVE,
                      reason='Recovery succeeded')])

    @mock.patch.object(nodem.Node, 'set_status')
    def test_node_recover_check_active(self, mock_status):
        node = nodem.Node('node1', PROFILE_ID, None)
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        node.status = 'ACTIVE'
        mock_check = self.patchobject(pb.Profile, 'check_object')
        mock_check.return_value = True
        action = mock.Mock(inputs={'check': True})

        res = node.do_recover(self.context, action)

        self.assertTrue(res)
        mock_check.assert_called_once_with(self.context, node)
        mock_status.assert_called_once_with(self.context, consts.NS_ACTIVE,
                                            reason='Recover: Node is ACTIVE.')

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'recover_object')
    def test_node_recover_check_error(self, mock_recover, mock_status):
        def set_status(*args, **kwargs):
            if args[1] == 'ACTIVE':
                node.physical_id = new_id
                node.data = {'recovery': 'RECREATE'}

        node = nodem.Node('node1', PROFILE_ID, '')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        new_id = '166db83b-b4a4-49ef-96a8-6c0fdd882d1a'
        mock_recover.return_value = new_id, True
        mock_status.side_effect = set_status
        mock_check = self.patchobject(pb.Profile, 'check_object')
        mock_check.return_value = False
        action = mock.Mock(inputs={'check': True})

        res = node.do_recover(self.context, action)

        self.assertTrue(res)
        mock_check.assert_called_once_with(self.context, node)
        mock_recover.assert_called_once_with(
            self.context, node, **action.inputs)
        self.assertEqual('node1', node.name)
        self.assertEqual(new_id, node.physical_id)
        self.assertEqual(PROFILE_ID, node.profile_id)
        mock_status.assert_has_calls([
            mock.call(self.context, 'RECOVERING',
                      reason='Recovery in progress'),
            mock.call(self.context, consts.NS_ACTIVE,
                      reason='Recovery succeeded',
                      physical_id=new_id,
                      data={'recovery': 'RECREATE'})])

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'recover_object')
    def test_node_recover_recreate(self, mock_recover, mock_status):
        def set_status(*args, **kwargs):
            if args[1] == 'ACTIVE':
                node.physical_id = new_id
                node.data = {'recovery': 'RECREATE'}

        node = nodem.Node('node1', PROFILE_ID, '', id='fake')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        new_id = '166db83b-b4a4-49ef-96a8-6c0fdd882d1a'
        mock_recover.return_value = new_id, True
        mock_status.side_effect = set_status
        mock_check = self.patchobject(pb.Profile, 'check_object')
        mock_check.return_value = False
        action = mock.Mock(
            outputs={}, inputs={'operation': [{'name': 'RECREATE'}],
                                'check': True})

        res = node.do_recover(self.context, action)

        self.assertTrue(res)
        mock_check.assert_called_once_with(self.context, node)
        mock_recover.assert_called_once_with(
            self.context, node, **action.inputs)
        self.assertEqual('node1', node.name)
        self.assertEqual(new_id, node.physical_id)
        self.assertEqual(PROFILE_ID, node.profile_id)
        mock_status.assert_has_calls([
            mock.call(self.context, 'RECOVERING',
                      reason='Recovery in progress'),
            mock.call(self.context, consts.NS_ACTIVE,
                      reason='Recovery succeeded',
                      physical_id=new_id,
                      data={'recovery': 'RECREATE'})])

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'recover_object')
    def test_node_recover_mult_rebuild(self, mock_recover, mock_status):
        def set_status(*args, **kwargs):
            if args[1] == 'ACTIVE':
                node.physical_id = new_id
                node.data = {'recovery': 'RECREATE'}

        node = nodem.Node('node1', PROFILE_ID, '', id='fake')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        new_id = '166db83b-b4a4-49ef-96a8-6c0fdd882d1a'
        mock_recover.return_value = new_id, True
        mock_status.side_effect = set_status
        mock_check = self.patchobject(pb.Profile, 'check_object')
        mock_check.return_value = False
        action = mock.Mock(
            outputs={}, inputs={'operation': [{'name': 'REBOOT'},
                                              {'name': 'REBUILD'}],
                                'check': True})

        res = node.do_recover(self.context, action)

        self.assertTrue(res)
        mock_check.assert_called_once_with(self.context, node)
        mock_recover.assert_called_once_with(
            self.context, node, **action.inputs)
        self.assertEqual('node1', node.name)
        self.assertEqual(new_id, node.physical_id)
        self.assertEqual(PROFILE_ID, node.profile_id)
        mock_status.assert_has_calls([
            mock.call(self.context, 'RECOVERING',
                      reason='Recovery in progress'),
            mock.call(self.context, consts.NS_ACTIVE,
                      reason='Recovery succeeded',
                      physical_id=new_id,
                      data={'recovery': 'RECREATE'})])

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'recover_object')
    def test_node_recover_check_exception(self, mock_recover, mock_status):
        def set_status(*args, **kwargs):
            if args[1] == 'ACTIVE':
                node.physical_id = new_id
                node.data = {'recovery': 'RECREATE'}

        node = nodem.Node('node1', PROFILE_ID, '')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        new_id = '166db83b-b4a4-49ef-96a8-6c0fdd882d1a'
        mock_recover.return_value = new_id, True
        mock_status.side_effect = set_status
        mock_check = self.patchobject(pb.Profile, 'check_object')
        mock_check.side_effect = exception.EResourceOperation(
            op='checking',
            type='server',
            id=node.physical_id,
            reason='Boom!'
        )
        action = mock.Mock(inputs={'operation': ['boom'],
                                   'check': True})

        res = node.do_recover(self.context, action)

        self.assertTrue(res)
        mock_check.assert_called_once_with(self.context, node)
        mock_recover.assert_called_once_with(
            self.context, node, **action.inputs)
        self.assertEqual('node1', node.name)
        self.assertEqual(new_id, node.physical_id)
        self.assertEqual(PROFILE_ID, node.profile_id)
        mock_status.assert_has_calls([
            mock.call(self.context, 'RECOVERING',
                      reason='Recovery in progress'),
            mock.call(self.context, consts.NS_ACTIVE,
                      reason='Recovery succeeded',
                      physical_id=new_id,
                      data={'recovery': 'RECREATE'})])

    @mock.patch.object(nodem.Node, 'set_status')
    @mock.patch.object(pb.Profile, 'recover_object')
    def test_node_recover_failed_recover(self, mock_recover, mock_status):
        node = nodem.Node('node1', PROFILE_ID, None)
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        mock_recover.return_value = node.physical_id, None
        action = mock.Mock(inputs={'operation': [{'name': 'RECREATE'}]})

        res = node.do_recover(self.context, action)

        self.assertFalse(res)
        mock_status.assert_has_calls([
            mock.call(self.context, 'RECOVERING',
                      reason='Recovery in progress'),
            mock.call(self.context, consts.NS_ERROR,
                      reason='Recovery failed')])

    def test_node_recover_no_physical_id(self):
        node = nodem.Node('node1', PROFILE_ID, None)
        action = mock.Mock()

        res = node.do_recover(self.context, action)

        self.assertFalse(res)

    @mock.patch.object(nodem.Node, 'set_status')
    def test_node_recover_operation_not_support(self, mock_set_status):
        node = nodem.Node('node1', PROFILE_ID, None)
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        action = mock.Mock(
            outputs={}, inputs={'operation': [{'name': 'foo'}]})

        res = node.do_recover(self.context, action)
        self.assertEqual({}, action.outputs)
        self.assertFalse(res)

    @mock.patch.object(nodem.Node, 'set_status')
    def test_node_operation(self, mock_set_status):
        node = nodem.Node('node1', PROFILE_ID, '')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        x_profile = mock.Mock()
        x_profile.handle_dance = mock.Mock(return_value=True)
        node.rt['profile'] = x_profile

        inputs = {'operation': 'dance', 'params': {'style': 'tango'}}
        res = node.do_operation(self.context, **inputs)

        self.assertTrue(res)
        mock_set_status.assert_has_calls([
            mock.call(self.context, consts.NS_OPERATING,
                      reason="Operation 'dance' in progress"),
            mock.call(self.context, consts.NS_ACTIVE,
                      reason="Operation 'dance' succeeded")
        ])
        x_profile.handle_dance.assert_called_once_with(node, style='tango')

    def test_node_operation_no_physical_id(self):
        node = nodem.Node('node1', PROFILE_ID, None)
        inputs = {'operation': 'dance', 'params': {'style': 'tango'}}

        res = node.do_operation(self.context, **inputs)

        self.assertFalse(res)

    @mock.patch.object(nodem.Node, 'set_status')
    def test_node_operation_failed_op(self, mock_set_status):
        node = nodem.Node('node1', PROFILE_ID, '')
        node.physical_id = 'd94d6333-82e6-4f87-b7ab-b786776df9d1'
        x_profile = mock.Mock()
        err = exception.EResourceOperation(
            op='dance', type='container', id='test_id', message='Boom')
        x_profile.handle_dance = mock.Mock(side_effect=err)
        node.rt['profile'] = x_profile

        inputs = {'operation': 'dance', 'params': {'style': 'tango'}}
        res = node.do_operation(self.context, **inputs)

        self.assertFalse(res)
        mock_set_status.assert_has_calls([
            mock.call(self.context, consts.NS_OPERATING,
                      reason="Operation 'dance' in progress"),
            mock.call(self.context, consts.NS_ERROR,
                      reason="Failed in dance container 'test_id': Boom.")
        ])
        x_profile.handle_dance.assert_called_once_with(node, style='tango')

    def test_run_workflow(self):
        node = nodem.Node('node1', PROFILE_ID, 'FAKE_CLUSTER')
        node.physical_id = 'FAKE_NODE'

        wfc = mock.Mock()
        wfc.workflow_find.return_value = None
        wfc.workflow_create = mock.Mock()
        wfc.execution_create = mock.Mock()
        x_profile = mock.Mock()
        x_profile.workflow = mock.Mock(return_value=wfc)
        node.rt['profile'] = x_profile

        options = {
            'workflow_name': 'foo',
            'inputs': {
                'definition': {
                    'bar': 'baz'
                },
                'FAKE_KEY1': 'FAKE_VALUE1',
                'FAKE_KEY2': 'FAKE_VALUE2',
            }
        }

        res = node.run_workflow(**options)

        self.assertTrue(res)
        x_profile.workflow.assert_called_once_with(node)
        wfc.workflow_find.assert_called_once_with('foo')
        wfc.workflow_create.assert_called_once_with(
            {'bar': 'baz'}, scope='private')

        final_dict = {
            'cluster_id': 'FAKE_CLUSTER',
            'node_id': 'FAKE_NODE',
            'FAKE_KEY1': 'FAKE_VALUE1',
            'FAKE_KEY2': 'FAKE_VALUE2',
        }
        final_inputs = jsonutils.dumps(final_dict)
        wfc.execution_create.assert_called_once_with(mock.ANY, mock.ANY)
        call_args, call_kwargs = wfc.execution_create.call_args
        self.assertEqual(call_args, ('foo', final_inputs))

    def test_run_workflow_no_physical_id(self):
        node = nodem.Node('node1', PROFILE_ID, 'FAKE_CLUSTER')
        node.physical_id = None

        res = node.run_workflow()

        self.assertFalse(res)

    def test_run_workflow_workflow_is_found(self):
        node = nodem.Node('node1', PROFILE_ID, 'FAKE_CLUSTER')
        node.physical_id = 'FAKE_NODE'

        wfc = mock.Mock()
        wfc.workflow_find.return_value = mock.Mock(definition={'bar': 'baz'})
        wfc.workflow_create = mock.Mock()
        wfc.execution_create = mock.Mock()
        x_profile = mock.Mock()
        x_profile.workflow = mock.Mock(return_value=wfc)
        node.rt['profile'] = x_profile

        options = {
            'workflow_name': 'foo',
            'inputs': {
                'FAKE_KEY1': 'FAKE_VALUE1',
                'FAKE_KEY2': 'FAKE_VALUE2',
            }
        }

        res = node.run_workflow(**options)

        self.assertTrue(res)
        x_profile.workflow.assert_called_once_with(node)
        wfc.workflow_find.assert_called_once_with('foo')
        self.assertEqual(0, wfc.workflow_create.call_count)
        final_dict = {
            'cluster_id': 'FAKE_CLUSTER',
            'node_id': 'FAKE_NODE',
            'FAKE_KEY1': 'FAKE_VALUE1',
            'FAKE_KEY2': 'FAKE_VALUE2',
        }
        final_inputs = jsonutils.dumps(final_dict)
        wfc.execution_create.assert_called_once_with(mock.ANY, mock.ANY)
        call_args, call_kwargs = wfc.execution_create.call_args
        self.assertEqual(call_args, ('foo', final_inputs))

    def test_run_workflow_failed_creation(self):
        node = nodem.Node('node1', PROFILE_ID, 'FAKE_CLUSTER')
        node.physical_id = 'FAKE_NODE'

        wfc = mock.Mock()
        wfc.workflow_find.return_value = None
        err = exception.InternalError(message='boom')
        wfc.workflow_create.side_effect = err
        wfc.execution_create = mock.Mock()
        x_profile = mock.Mock()
        x_profile.workflow = mock.Mock(return_value=wfc)
        node.rt['profile'] = x_profile

        options = {
            'workflow_name': 'foo',
            'inputs': {
                'definition': {'bar': 'baz'},
                'FAKE_KEY1': 'FAKE_VALUE1',
                'FAKE_KEY2': 'FAKE_VALUE2',
            }
        }

        ex = self.assertRaises(exception.EResourceOperation,
                               node.run_workflow,
                               **options)

        self.assertEqual("Failed in executing workflow 'foo': boom.",
                         six.text_type(ex))
        x_profile.workflow.assert_called_once_with(node)
        wfc.workflow_find.assert_called_once_with('foo')
        wfc.workflow_create.assert_called_once_with(
            {'bar': 'baz'}, scope='private')
        self.assertEqual(0, wfc.execution_create.call_count)

    def test_run_workflow_failed_execution(self):
        node = nodem.Node('node1', PROFILE_ID, 'FAKE_CLUSTER')
        node.physical_id = 'FAKE_NODE'

        wfc = mock.Mock()
        wfc.workflow_find.return_value = None
        wfc.workflow_create = mock.Mock()
        err = exception.InternalError(message='boom')
        wfc.execution_create.side_effect = err
        x_profile = mock.Mock()
        x_profile.workflow = mock.Mock(return_value=wfc)
        node.rt['profile'] = x_profile

        options = {
            'workflow_name': 'foo',
            'inputs': {
                'definition': {'bar': 'baz'},
                'FAKE_KEY1': 'FAKE_VALUE1',
                'FAKE_KEY2': 'FAKE_VALUE2',
            }
        }

        ex = self.assertRaises(exception.EResourceOperation,
                               node.run_workflow,
                               **options)

        self.assertEqual("Failed in executing workflow 'foo': boom.",
                         six.text_type(ex))
        x_profile.workflow.assert_called_once_with(node)
        wfc.workflow_find.assert_called_once_with('foo')
        wfc.workflow_create.assert_called_once_with(
            {'bar': 'baz'}, scope='private')
        final_dict = {
            'cluster_id': 'FAKE_CLUSTER',
            'node_id': 'FAKE_NODE',
            'FAKE_KEY1': 'FAKE_VALUE1',
            'FAKE_KEY2': 'FAKE_VALUE2',
        }
        final_inputs = jsonutils.dumps(final_dict)
        wfc.execution_create.assert_called_once_with(mock.ANY, mock.ANY)
        call_args, call_kwargs = wfc.execution_create.call_args
        self.assertEqual(call_args, ('foo', final_inputs))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging

from senlin.common import context
from senlin.engine.notifications import base
from senlin import objects
from senlin.rpc import client as rpc_client

LOG = logging.getLogger(__name__)


class NovaNotificationEndpoint(base.Endpoints):

    VM_FAILURE_EVENTS = {
        'compute.instance.pause.end': 'PAUSE',
        'compute.instance.power_off.end': 'POWER_OFF',
        'compute.instance.rebuild.error': 'REBUILD',
        'compute.instance.shutdown.end': 'SHUTDOWN',
        'compute.instance.soft_delete.end': 'SOFT_DELETE',
    }

    def __init__(self, project_id, engine_id, recover_action):
        self.filter_rule = messaging.NotificationFilter(
            publisher_id='^compute.*',
            event_type='^compute\.instance\..*',
            context={'project_id': '^%s$' % project_id})
        self.project_id = project_id
        self.engine_id = engine_id
        self.rpc = rpc_client.EngineClient()
        self.recover_action = recover_action
        self.exchange = cfg.CONF.health_manager.nova_control_exchange
        self.target = messaging.Target(topic='versioned_notifications',
                                       exchange=self.exchange)

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        meta = payload['metadata']
        cluster_id = meta.get('cluster_id')
        if not cluster_id:
            return

        if event_type not in self.VM_FAILURE_EVENTS:
            return

        ctx = context.get_service_context(project=self.project_id,
                                          user=payload['user_id'])
        enabled = self._check_registry_status(ctx, self.engine_id, cluster_id)
        if enabled is False:
            return

        params = {
            'event': self.VM_FAILURE_EVENTS[event_type],
            'state': payload.get('state', 'Unknown'),
            'instance_id': payload.get('instance_id', 'Unknown'),
            'timestamp': metadata['timestamp'],
            'publisher': publisher_id,
            'operation': self.recover_action['operation'],
        }
        node_id = meta.get('cluster_node_id')
        if node_id:
            LOG.info("Requesting node recovery: %s", node_id)
            req = objects.NodeRecoverRequest(identity=node_id,
                                             params=params)
            self.rpc.call(ctx, 'node_recover', req)

    def warn(self, ctxt, publisher_id, event_type, payload, metadata):
        meta = payload.get('metadata', {})
        if meta.get('cluster_id') == self.cluster_id:
            LOG.warning("publisher=%s", publisher_id)
            LOG.warning("event_type=%s", event_type)

    def debug(self, ctxt, publisher_id, event_type, payload, metadata):
        meta = payload.get('metadata', {})
        if meta.get('cluster_id') == self.cluster_id:
            LOG.debug("publisher=%s", publisher_id)
            LOG.debug("event_type=%s", event_type)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

import mock
from oslo_config import cfg

from senlin.drivers.os import keystone_v3 as kv3
from senlin.drivers import sdk
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


@mock.patch.object(sdk, 'create_connection')
class TestKeystoneV3(base.SenlinTestCase):

    def setUp(self):
        super(TestKeystoneV3, self).setUp()

        self.ctx = utils.dummy_context()
        self.conn = mock.Mock()

    def test_init(self, mock_create):
        mock_create.return_value = self.conn
        kc = kv3.KeystoneClient({'k': 'v'})

        mock_create.assert_called_once_with({'k': 'v'})
        self.assertEqual(self.conn, kc.conn)
        self.assertEqual(self.conn.session, kc.session)

    def test_trust_get_by_trustor(self, mock_create):
        trust1 = mock.Mock()
        trust1.trustee_user_id = 'USER_A_ID'
        trust1.project_id = 'PROJECT_ID_1'

        trust2 = mock.Mock()
        trust2.trustee_user_id = 'USER_B_ID'
        trust2.project_id = 'PROJECT_ID_1'

        trust3 = mock.Mock()
        trust3.trustee_user_id = 'USER_A_ID'
        trust3.project_id = 'PROJECT_ID_2'

        self.conn.identity.trusts.return_value = [trust1, trust2, trust3]
        mock_create.return_value = self.conn
        kc = kv3.KeystoneClient({'k': 'v'})

        # no trustee/project filter, matching 1st
        res = kc.trust_get_by_trustor('USER_A')
        self.assertEqual(trust1, res)

        # trustee specified, matching 2nd
        res = kc.trust_get_by_trustor('USER_A', 'USER_B_ID')
        self.assertEqual(trust2, res)

        # project specified, matching 3rd
        res = kc.trust_get_by_trustor('USER_A', project='PROJECT_ID_2')
        self.assertEqual(trust3, res)

        # both trustee and project specified, matching 3rd
        res = kc.trust_get_by_trustor('USER_A', 'USER_A_ID', 'PROJECT_ID_2')
        self.assertEqual(trust3, res)

        # No matching record found
        res = kc.trust_get_by_trustor('USER_A', 'USER_C_ID')
        self.assertIsNone(res)

        get_calls = [mock.call(trustor_user_id='USER_A')]
        self.conn.identity.trusts.assert_has_calls(get_calls * 5)

    def test_trust_create(self, mock_create):
        self.conn.identity.create_trust.return_value = 'new_trust'
        mock_create.return_value = self.conn
        kc = kv3.KeystoneClient({'k': 'v'})

        # default
        res = kc.trust_create('ID_JOHN', 'ID_DOE', 'PROJECT_ID')

        self.assertEqual('new_trust', res)
        self.conn.identity.create_trust.assert_called_once_with(
            trustor_user_id='ID_JOHN', trustee_user_id='ID_DOE',
            project_id='PROJECT_ID', impersonation=True,
            allow_redelegation=True, roles=[])
        self.conn.reset_mock()

        # with roles
        res = kc.trust_create('ID_JOHN', 'ID_DOE', 'PROJECT_ID',
                              ['r1', 'r2'])

        self.assertEqual('new_trust', res)
        self.conn.identity.create_trust.assert_called_once_with(
            trustor_user_id='ID_JOHN', trustee_user_id='ID_DOE',
            project_id='PROJECT_ID', impersonation=True,
            allow_redelegation=True,
            roles=[{'name': 'r1'}, {'name': 'r2'}])
        self.conn.reset_mock()

        # impersonation
        res = kc.trust_create('ID_JOHN', 'ID_DOE', 'PROJECT_ID',
                              impersonation=False)

        self.assertEqual('new_trust', res)
        self.conn.identity.create_trust.assert_called_once_with(
            trustor_user_id='ID_JOHN', trustee_user_id='ID_DOE',
            project_id='PROJECT_ID', impersonation=False,
            allow_redelegation=True, roles=[])
        self.conn.reset_mock()

    @mock.patch.object(sdk, 'authenticate')
    def test_get_token(self, mock_auth, mock_create):
        access_info = {'token': '123', 'user_id': 'abc', 'project_id': 'xyz'}
        mock_auth.return_value = access_info

        token = kv3.KeystoneClient.get_token(key='value')

        mock_auth.assert_called_once_with(key='value')
        self.assertEqual('123', token)

    @mock.patch.object(sdk, 'authenticate')
    def test_get_user_id(self, mock_auth, mock_create):
        access_info = {'token': '123', 'user_id': 'abc', 'project_id': 'xyz'}
        mock_auth.return_value = access_info

        user_id = kv3.KeystoneClient.get_user_id(key='value')

        mock_auth.assert_called_once_with(key='value')
        self.assertEqual('abc', user_id)

    def test_get_service_credentials(self, mock_create):
        cfg.CONF.set_override('auth_url', 'FAKE_URL', group='authentication')
        cfg.CONF.set_override('service_username', 'FAKE_USERNAME',
                              group='authentication')
        cfg.CONF.set_override('service_password', 'FAKE_PASSWORD',
                              group='authentication')
        cfg.CONF.set_override('service_project_name', 'FAKE_PROJECT',
                              group='authentication')
        cfg.CONF.set_override('service_user_domain', 'FAKE_DOMAIN_1',
                              group='authentication')
        cfg.CONF.set_override('service_project_domain', 'FAKE_DOMAIN_2',
                              group='authentication')
        expected = {
            'auth_url': 'FAKE_URL',
            'username': 'FAKE_USERNAME',
            'password': 'FAKE_PASSWORD',
            'project_name': 'FAKE_PROJECT',
            'user_domain_name': 'FAKE_DOMAIN_1',
            'project_domain_name': 'FAKE_DOMAIN_2'
        }

        actual = kv3.KeystoneClient.get_service_credentials()

        self.assertEqual(expected, actual)

        new_expected = copy.copy(expected)
        new_expected['key1'] = 'value1'
        new_expected['password'] = 'NEW_PASSWORD'

        actual = kv3.KeystoneClient.get_service_credentials(
            key1='value1', password='NEW_PASSWORD')

        self.assertEqual(new_expected, actual)

    def test_validate_regions(self, mock_create):
        self.conn.identity.regions.return_value = [
            {'id': 'R1', 'parent_region_id': None},
            {'id': 'R2', 'parent_region_id': None},
            {'id': 'R3', 'parent_region_id': 'R1'},
        ]
        mock_create.return_value = self.conn

        kc = kv3.KeystoneClient({'k': 'v'})

        res = kc.validate_regions(['R1', 'R4'])

        self.assertIn('R1', res)
        self.assertNotIn('R4', res)

        res = kc.validate_regions([])
        self.assertEqual([], res)

    def test_get_senlin_endpoint(self, mock_create):
        cfg.CONF.set_override('default_region_name', 'RegionN')
        self.conn.session.get_endpoint.return_value = 'http://web.com:1234/v1'
        mock_create.return_value = self.conn
        kc = kv3.KeystoneClient({'k': 'v'})

        res = kc.get_senlin_endpoint()

        self.assertEqual('http://web.com:1234/v1', res)
        self.conn.session.get_endpoint.assert_called_once_with(
            service_type='clustering', interface='public', region='RegionN')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging

from senlin.common import context
from senlin.engine.notifications import base
from senlin import objects
from senlin.rpc import client as rpc_client

LOG = logging.getLogger(__name__)


class HeatNotificationEndpoint(base.Endpoints):

    STACK_FAILURE_EVENTS = {
        'orchestration.stack.delete.end': 'DELETE',
    }

    def __init__(self, project_id, engine_id, recover_action):
        self.filter_rule = messaging.NotificationFilter(
            publisher_id='^orchestration.*',
            event_type='^orchestration\.stack\..*',
            context={'project_id': '^%s$' % project_id})
        self.project_id = project_id
        self.engine_id = engine_id
        self.rpc = rpc_client.EngineClient()
        self.recover_action = recover_action
        self.exchange = cfg.CONF.health_manager.heat_control_exchange
        self.target = messaging.Target(topic='notifications',
                                       exchange=self.exchange)

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        if event_type not in self.STACK_FAILURE_EVENTS:
            return

        tags = payload['tags']
        if tags is None or tags == []:
            return

        cluster_id = None
        node_id = None
        for tag in tags:
            if cluster_id is None:
                start = tag.find('cluster_id')
                if start == 0 and tag[11:]:
                    cluster_id = tag[11:]
            if node_id is None:
                start = tag.find('cluster_node_id')
                if start == 0:
                    node_id = tag[16:]

        if cluster_id is None or node_id is None:
            return

        ctx = context.get_service_context(project=self.project_id,
                                          user=payload['user_identity'])
        enabled = self._check_registry_status(ctx, self.engine_id, cluster_id)
        if enabled is False:
            return

        params = {
            'event': self.STACK_FAILURE_EVENTS[event_type],
            'state': payload.get('state', 'Unknown'),
            'stack_id': payload.get('stack_identity', 'Unknown'),
            'timestamp': metadata['timestamp'],
            'publisher': publisher_id,
            'operation': self.recover_action['operation'],
        }
        LOG.info("Requesting stack recovery: %s", node_id)
        req = objects.NodeRecoverRequest(identity=node_id, params=params)
        self.rpc.call(ctx, 'node_recover', req)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Senlin common internal object model"""
import re

from oslo_utils import versionutils
from oslo_versionedobjects import base
from oslo_versionedobjects import fields as base_fields

from senlin.common.i18n import _
from senlin import objects

VersionedObjectDictCompat = base.VersionedObjectDictCompat
VersionedObjectSerializer = base.VersionedObjectSerializer


class SenlinObject(base.VersionedObject):
    """Base class for senlin objects.

    This is the base class for all objects that can be remoted or instantiated
    via RPC. Simply defining a sub-class of this class would make it remotely
    instantiatable. Objects should implement the "get" class method and the
    "save" object method.
    """
    OBJ_SERIAL_NAMESPACE = 'senlin_object'
    OBJ_PROJECT_NAMESPACE = 'senlin'
    BASE_VERSION = '1.0'
    VERSION = '1.0'

    # list of version maps from api request version to object version
    # higher api versions after lower api versions. e.g.
    # {'1.2': '1.0', '1.4': '1.1'}
    VERSION_MAP = {}

    @staticmethod
    def _from_db_object(context, obj, db_obj):
        if db_obj is None:
            return None
        for field in obj.fields:
            if field == 'metadata':
                obj['metadata'] = db_obj['meta_data']
            else:
                obj[field] = db_obj[field]

        obj._context = context
        obj.obj_reset_changes()

        return obj

    @staticmethod
    def _transpose_metadata(values):
        """Utility function to translate metadata field."""
        if 'metadata' in values:
            value = values.pop('metadata')
            values['meta_data'] = value
        return values

    @classmethod
    def to_json_schema(cls):
        obj_name = cls.obj_name()
        schema = {
            '$schema': 'http://json-schema.org/draft-04/schema#',
            'title': obj_name,
        }

        schema.update(base_fields.Object(obj_name).get_schema())
        dataf = cls.OBJ_SERIAL_NAMESPACE + ".data"
        schema["properties"][dataf]["additionalProperties"] = False
        return schema

    @classmethod
    def obj_class_from_name(cls, objname, objver=None):
        if objver is None:
            objver = cls.VERSION
        return super(SenlinObject, cls).obj_class_from_name(objname, objver)

    @classmethod
    def find_version(cls, context):
        match = re.match(r"^([1-9]\d*)\.([1-9]\d*|0)$", context.api_version)
        req_major = int(match.group(1))
        req_minor = int(match.group(2))
        # base version is '1.0'
        matched_version = cls.BASE_VERSION
        for api_ver in sorted(cls.VERSION_MAP.keys()):
            match = re.match(r"^([1-9]\d*)\.([1-9]\d*|0)$", api_ver)
            api_major = int(match.group(1))
            api_minor = int(match.group(2))
            if (api_major, api_minor) <= (req_major, req_minor):
                matched_version = cls.VERSION_MAP[api_ver]
            else:
                break

        return matched_version

    @classmethod
    def normalize_req(cls, name, req, key=None):
        result = {
            cls.OBJ_SERIAL_NAMESPACE + '.version': cls.VERSION,
            cls.OBJ_SERIAL_NAMESPACE + '.namespace': cls.OBJ_PROJECT_NAMESPACE,
            cls.OBJ_SERIAL_NAMESPACE + '.name': name,
        }
        if key is not None:
            if key not in req:
                raise ValueError(_("Request body missing '%s' key.") % key)

            result[cls.OBJ_SERIAL_NAMESPACE + '.data'] = {
                key: {
                    cls.OBJ_SERIAL_NAMESPACE + '.version': cls.VERSION,
                    cls.OBJ_SERIAL_NAMESPACE + '.namespace':
                        cls.OBJ_PROJECT_NAMESPACE,
                    cls.OBJ_SERIAL_NAMESPACE + '.name': name + 'Body',
                    cls.OBJ_SERIAL_NAMESPACE + '.data': req[key]
                }
            }
        else:
            result[cls.OBJ_SERIAL_NAMESPACE + '.data'] = req

        return result


class SenlinObjectRegistry(base.VersionedObjectRegistry):

    notification_classes = []

    def registration_hook(self, cls, index):
        """Callback for object registration.

        When an object is registered, this function will be called for
        maintaining senlin.objects.$OBJECT as the highest-versioned
        implementation of a given object.
        """
        version = versionutils.convert_version_to_tuple(cls.VERSION)
        if not hasattr(objects, cls.obj_name()):
            setattr(objects, cls.obj_name(), cls)
        else:
            curr_version = versionutils.convert_version_to_tuple(
                getattr(objects, cls.obj_name()).VERSION)
            if version >= curr_version:
                setattr(objects, cls.obj_name(), cls)

    @classmethod
    def register_notification(cls, notification_cls):
        """Register a class as concrete notification.

        This is used only to register concrete notification or payload
        classes. Do NOT register base classes intended for inheritance only.
        """
        cls.register_if(False)(notification_cls)
        cls.notification_classes.append(notification_cls)
        return notification_cls



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Action dependency object."""

from senlin.db import api as db_api
from senlin.objects import base
from senlin.objects import fields


@base.SenlinObjectRegistry.register
class Dependency(base.SenlinObject, base.VersionedObjectDictCompat):
    """Senlin action dependency object."""

    fields = {
        'id': fields.UUIDField(),
        'depended': fields.UUIDField(),
        'dependent': fields.UUIDField(),
    }

    @classmethod
    def create(cls, context, depended, dependent):
        return db_api.dependency_add(context, depended, dependent)

    @classmethod
    def get_depended(cls, context, action_id):
        return db_api.dependency_get_depended(context, action_id)

    @classmethod
    def get_dependents(cls, context, action_id):
        return db_api.dependency_get_dependents(context, action_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_utils import timeutils
from senlin.api.common import util
from senlin.api.common import wsgi
from senlin.common import exception
from senlin.objects import service as service_obj

CONF = cfg.CONF


class ServiceController(wsgi.Controller):
    """WSGI controller for Services in Senlin v1 API."""

    REQUEST_SCOPE = 'services'

    @util.policy_enforce
    def index(self, req):
        if not req.context.is_admin:
            raise exception.Forbidden()
        now = timeutils.utcnow(with_timezone=True)
        _services = service_obj.Service.get_all(req.context)
        svcs = []
        for svc in _services:
            updated_at = svc.updated_at
            delta = now - (svc.updated_at or svc.created_at)
            delta_sec = delta.total_seconds()
            alive = abs(delta_sec) <= CONF.service_down_time
            art = (alive and "up") or "down"
            active = 'enabled'
            if svc.disabled:
                active = 'disabled'
            if updated_at:
                updated_at = timeutils.normalize_time(updated_at)
            ret_fields = {'id': svc.id, 'host': svc.host,
                          'binary': svc.binary, 'topic': svc.topic,
                          'disabled_reason': svc.disabled_reason,
                          'status': active, 'state': art,
                          'updated_at': updated_at}
            svcs.append(ret_fields)
        return {'services': svcs}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_context import context as oslo_context
from oslo_log import log as logging
import six
import tenacity

from senlin.common import context as senlin_context
from senlin.common import exception
from senlin.drivers import base as driver_base
from senlin.objects import credential as co

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

RETRY_ATTEMPTS = 3
RETRY_INITIAL_DELAY = 1
RETRY_BACKOFF = 1
RETRY_MAX = 3


class Message(object):
    """Zaqar message type of notification."""
    def __init__(self, queue_name, **kwargs):
        self.user = kwargs.get('user', '')
        self.project = kwargs.get('project', '')
        self.domain = kwargs.get('domain', '')

        self.queue_name = queue_name

        self._zaqarclient = None
        self._keystoneclient = None

    def zaqar(self):
        if self._zaqarclient is not None:
            return self._zaqarclient
        params = self._build_conn_params(self.user, self.project)
        self._zaqarclient = driver_base.SenlinDriver().message(params)
        return self._zaqarclient

    def _build_conn_params(self, user, project):
        """Build connection params for specific user and project.

        :param user: The ID of the user for which a trust will be used.
        :param project: The ID of the project for which a trust will be used.
        :returns: A dict containing the required parameters for connection
                  creation.
        """
        service_creds = senlin_context.get_service_credentials()
        params = {
            'username': service_creds.get('username'),
            'password': service_creds.get('password'),
            'auth_url': service_creds.get('auth_url'),
            'user_domain_name': service_creds.get('user_domain_name')
        }

        cred = co.Credential.get(oslo_context.get_current(), user, project)
        if cred is None:
            raise exception.TrustNotFound(trustor=user)
        params['trust_id'] = cred.cred['openstack']['trust']

        return params

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(exception.EResourceCreation),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def post_lifecycle_hook_message(self, lifecycle_action_token, node_id,
                                    resource_id, lifecycle_transition_type):
        message_list = [{
            "ttl": CONF.notification.ttl,
            "body": {
                "lifecycle_action_token": lifecycle_action_token,
                "node_id": node_id,
                "resource_id": resource_id,
                "lifecycle_transition_type": lifecycle_transition_type
            }
        }]
        try:
            if not self.zaqar().queue_exists(self.queue_name):
                kwargs = {
                    "_max_messages_post_size":
                        CONF.notification.max_message_size,
                    "description": "Senlin lifecycle hook notification",
                    "name": self.queue_name
                }
                self.zaqar().queue_create(**kwargs)

            return self.zaqar().message_post(self.queue_name, message_list)
        except exception.InternalError as ex:
            raise exception.EResourceCreation(
                type='queue',
                message=six.text_type(ex))
